"""Test the Store interface

Many methods simply proxy to config/metadata/data store implementations, but there is some
cross-coordination and there are some convenience methods implemented at this layer.
"""
import numpy.testing
from pytest import fixture
from smif.data_layer import Store
from smif.data_layer.memory_interface import (MemoryConfigStore,
                                              MemoryDataStore,
                                              MemoryMetadataStore)


@fixture
def store():
    """Store fixture
    """
    # implement each part using the memory classes, simpler than mocking
    # each other implementation of a part is tested fully by e.g. test_config_store.py
    return Store(
        config_store=MemoryConfigStore(),
        metadata_store=MemoryMetadataStore(),
        data_store=MemoryDataStore()
    )


class TestStoreConfig():
    def test_model_runs(self, store, minimal_model_run):
        # write
        store.write_model_run(minimal_model_run)
        # read all
        assert store.read_model_runs() == [minimal_model_run]
        # read one
        assert store.read_model_run(minimal_model_run['name']) == minimal_model_run
        # update
        store.update_model_run(minimal_model_run['name'], minimal_model_run)
        # delete
        store.delete_model_run(minimal_model_run['name'])
        assert store.read_model_runs() == []

    def test_read_model_run_sorted(self, store, minimal_model_run):
        y_model_run = {'name': 'y'}
        z_model_run = {'name': 'z'}

        store.write_model_run(minimal_model_run)
        store.write_model_run(z_model_run)
        store.write_model_run(y_model_run)

        expected = [minimal_model_run, y_model_run, z_model_run]
        assert store.read_model_runs() == expected

    def test_sos_models(self, store, get_sos_model, get_sector_model,
                        energy_supply_sector_model, sample_scenarios):
        # write
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        for scenario in sample_scenarios:
            store.write_scenario(scenario)
        store.write_sos_model(get_sos_model)
        # read all
        assert store.read_sos_models() == [get_sos_model]
        # read one
        assert store.read_sos_model(get_sos_model['name']) == get_sos_model
        # update
        store.update_sos_model(get_sos_model['name'], get_sos_model)
        # delete
        store.delete_sos_model(get_sos_model['name'])
        assert store.read_sos_models() == []

    def test_models(self, store, get_sector_model, sample_dimensions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        # write
        store.write_model(get_sector_model)
        # read all
        assert store.read_models() == [get_sector_model]
        # read one
        assert store.read_model(get_sector_model['name']) == get_sector_model
        # update
        store.update_model(get_sector_model['name'], get_sector_model)
        # delete
        store.delete_model(get_sector_model['name'])
        assert store.read_models() == []
        # teardown
        for dim in sample_dimensions:
            store.delete_dimension(dim['name'])

    def test_models_skip_coords(self, store, get_sector_model, get_sector_model_no_coords):
        # write
        store.write_model(get_sector_model)
        # read all
        assert store.read_models(skip_coords=True) == [get_sector_model_no_coords]
        # read one
        assert store.read_model(get_sector_model['name'], skip_coords=True) == \
            get_sector_model_no_coords
        # update
        store.update_model(get_sector_model['name'], get_sector_model)
        # delete
        store.delete_model(get_sector_model['name'])
        assert store.read_models() == []

    def test_scenarios(self, store, scenario, sample_dimensions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        # write
        store.write_scenario(scenario)
        # read all
        assert store.read_scenarios() == [scenario]
        # read one
        assert store.read_scenario(scenario['name']) == scenario
        # update
        store.update_scenario(scenario['name'], scenario)
        # delete
        store.delete_scenario(scenario['name'])
        assert store.read_scenarios() == []
        # teardown
        for dim in sample_dimensions:
            store.delete_dimension(dim['name'])

    def test_scenarios_skip_coords(self, store, scenario, scenario_no_coords):
        # write
        store.write_scenario(scenario)
        # read all
        assert store.read_scenarios(skip_coords=True) == [scenario_no_coords]
        # read one
        assert store.read_scenario(scenario['name'], skip_coords=True) == \
            scenario_no_coords
        # update
        store.update_scenario(scenario['name'], scenario)
        # delete
        store.delete_scenario(scenario['name'])
        assert store.read_scenarios() == []

    def test_scenario_variants(self, store, scenario):
        scenario_name = scenario['name']
        old_variant = scenario['variants'][0]
        # write
        store.write_scenario(scenario)
        new_variant = {
            'name': 'high',
            'description': 'Mortality (High)',
            'data': {
                'mortality': 'mortality_high.csv'
            }
        }
        store.write_scenario_variant(scenario_name, new_variant)
        # read all
        both = store.read_scenario_variants(scenario_name)
        assert both == [old_variant, new_variant] or both == [new_variant, old_variant]
        # read one
        assert store.read_scenario_variant(scenario_name, old_variant['name']) == old_variant
        # update
        store.update_scenario_variant(scenario_name, old_variant['name'], old_variant)
        # delete
        store.delete_scenario_variant(scenario_name, old_variant['name'])
        assert store.read_scenario_variants(scenario_name) == [new_variant]

    def test_narratives(self, store, get_sos_model):
        store.write_sos_model(get_sos_model)
        expected = get_sos_model['narratives'][0]
        # read one
        assert store.read_narrative(get_sos_model['name'], expected['name']) == expected

    def test_strategies(self, store, strategies):
        model_run_name = 'test_modelrun'
        # write
        store.write_strategies(model_run_name, strategies)
        # read
        assert store.read_strategies(model_run_name) == strategies


class TestStoreMetadata():
    def test_units(self, store, unit_definitions):
        # write
        store.write_unit_definitions(unit_definitions)
        # read
        assert store.read_unit_definitions() == unit_definitions

    def test_dimensions(self, store, dimension):
        # write
        store.write_dimension(dimension)
        # read all
        assert store.read_dimensions() == [dimension]
        # read one
        assert store.read_dimension(dimension['name']) == dimension
        # update
        store.update_dimension(dimension['name'], dimension)
        # delete
        store.delete_dimension(dimension['name'])
        assert store.read_dimensions() == []


class TestStoreData():
    def test_scenario_variant_data(self, store, sample_dimensions, scenario,
                                   sample_scenario_data):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_scenario(scenario)
        # pick out single sample
        key = next(iter(sample_scenario_data))
        scenario_name, variant_name, variable = key
        scenario_variant_data = sample_scenario_data[key]
        # write
        store.write_scenario_variant_data(
            scenario_name, variant_name, scenario_variant_data
        )
        # read
        actual = store.read_scenario_variant_data(
            scenario_name, variant_name, variable
        )
        assert actual == scenario_variant_data

    def test_narrative_variant_data(self, store, sample_dimensions, get_sos_model,
                                    get_sector_model, energy_supply_sector_model,
                                    sample_narrative_data):
        # Setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        # pick out single sample
        key = (
            'energy',
            'technology',
            'high_tech_dsm',
            'smart_meter_savings'
        )
        sos_model_name, narrative_name, variant_name, param_name = key
        narrative_variant_data = sample_narrative_data[key]
        # write
        store.write_narrative_variant_data(
            sos_model_name, narrative_name, variant_name, narrative_variant_data)
        # read
        actual = store.read_narrative_variant_data(
            sos_model_name, narrative_name, variant_name, param_name)
        assert actual == narrative_variant_data

    def test_model_parameter_default(self, store, get_multidimensional_param,
                                     get_sector_model, sample_dimensions):
        param_data = get_multidimensional_param
        for dim in sample_dimensions:
            store.write_dimension(dim)
        for dim in param_data.dims:
            store.write_dimension({'name': dim, 'elements': param_data.dim_elements(dim)})
        get_sector_model['parameters'] = [param_data.spec.as_dict()]
        store.write_model(get_sector_model)
        # write
        store.write_model_parameter_default(
            get_sector_model['name'], param_data.name, param_data)
        # read
        actual = store.read_model_parameter_default(get_sector_model['name'], param_data.name)
        assert actual == param_data

    def test_interventions(self, store, sample_dimensions, get_sector_model, interventions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_model(get_sector_model)
        # write
        store.write_interventions(get_sector_model['name'], interventions)
        # read
        assert store.read_interventions(get_sector_model['name']) == interventions

    def test_initial_conditions(self, store, sample_dimensions, initial_conditions,
                                get_sos_model, get_sector_model, energy_supply_sector_model,
                                minimal_model_run):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model_run(minimal_model_run)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        # write
        store.write_initial_conditions(get_sector_model['name'], initial_conditions)
        # read
        assert store.read_initial_conditions(get_sector_model['name']) == initial_conditions
        # read all for a model run
        actual = store.read_all_initial_conditions(minimal_model_run['name'])
        assert actual == initial_conditions

    def test_state(self, store, state):
        # write
        store.write_state(state, 'model_run_name', 0, 0)
        # read
        assert store.read_state('model_run_name', 0, 0) == state

    def test_conversion_coefficients(self, store, conversion_coefficients):
        # write
        store.write_coefficients(
            'source_dim', 'sink_dim', conversion_coefficients)
        # read
        numpy.testing.assert_equal(
            store.read_coefficients('source_dim', 'sink_dim'),
            conversion_coefficients
        )

    def test_results(self, store, sample_results):
        # write
        store.write_results(sample_results, 'model_run_name', 'model_name', 0)
        # read
        spec = sample_results.spec
        assert store.read_results('model_run_name', 'model_name', spec, 0) == sample_results
        # check
        assert store.available_results('model_run_name') == \
            [(0, None, 'model_name', spec.name)]

    def test_warm_start(self, store, sample_results):
        assert store.prepare_warm_start('test_model_run') is None
        timestep = 2020
        store.write_results(sample_results, 'test_model_run', 'model_name', timestep)
        assert store.prepare_warm_start('test_model_run') == timestep



"""Implements procedures and tools to run a system-of-systems model
Provides methods for copying a sample project into a local directory
and for or running a system-of-systems model.

Provides a ModelRunScheduler that can run modelruns as a subprocess.

Example
-------
Copy a project folder::

>>> copy_project_folder('/projects/smif/')

Run a single system-of-systems model::

>>> execute_model_run('energy_supply_demand', store)

Use the ModelRunScheduler to run a system-of-systems model::

    >>> my_scheduler = ModelRunScheduler()
    >>> my_scheduler.add(
            'energy_supply_demand',
            {
                'directory': '/projects/smif'
            }
        )
    >>> my_scheduler.status('energy_supply_demand')
    {
        'message': 'running'
    }
    >>> my_scheduler.status('energy_supply_demand')
    {
    "err": "",
    "output": "20102010\\n2015\\n2015\\nModel run '20170918_energy_water' complete\\n",
    "status": "done"
    }
"""

# import classes for access like ::
#         from smif.controller import ModelRunScheduler
from smif.controller.scheduler import ModelRunScheduler
from smif.controller.execute import execute_model_run
from smif.controller.setup import copy_project_folder
from smif.controller.modelrun import ModelRunner

# Define what should be imported as * ::
#         from smif.controller import *
__all__ = ['ModelRunner', 'ModelRunScheduler', 'execute_model_run', 'copy_project_folder']



"""HTTP API endpoint
"""
from collections import defaultdict

import dateutil.parser
import smif
from flask import current_app, jsonify, request
from flask.views import MethodView
from smif.exception import (SmifDataError, SmifDataInputError,
                            SmifDataNotFoundError, SmifException,
                            SmifValidationError)


class SmifAPI(MethodView):
    """Implement operations for Smif
    """
    def get(self, key):
        """Get smif details
        version: GET /api/v1/smif/version
        """
        if key == 'version':
            data = smif.__version__
        else:
            data = {}
            data['version'] = smif.__version__

        response = jsonify({
            'data': data,
            'error': {}
        })
        return response


class ModelRunAPI(MethodView):
    """Implement CRUD operations for model_run configuration data
    """
    def get(self, model_run_name=None, action=None):
        """Get model_runs
        all: GET /api/v1/model_runs/
        one: GET /api/vi/model_runs/name
        """
        data_interface = current_app.config.data_interface

        try:
            if action is None:
                if model_run_name is None:

                    model_runs = data_interface.read_model_runs()

                    if 'status' in request.args.keys():
                        # filtered: GET /api/v1/model_runs?status=done
                        data = []
                        for model_run in model_runs:
                            status = current_app.config.scheduler.get_status(model_run['name'])
                            if status['status'] == request.args['status']:
                                data.append(model_run)
                    else:
                        # all: GET /api/v1/model_runs/
                        data = []
                        data = model_runs
                else:
                    # one: GET /api/vi/model_runs/name
                    data = {}
                    data = data_interface.read_model_run(model_run_name)
            elif action == 'status':
                # action: GET /api/vi/model_runs/name/status
                data = {}
                data = current_app.config.scheduler.get_status(model_run_name)

            response = jsonify({
                'data': data,
                'error': {}
            })
        except SmifException as err:
            response = jsonify({
                'data': data,
                'error': parse_exceptions(err)
            })

        return response

    def post(self, model_run_name=None, action=None):
        """
        Create a model_run:
        - POST /api/v1/model_runs

        Perform an operation on a model_run
        - POST /api/v1/model_runs/<model_run_name>/<action>

        Available actions are
        - start: Start the model_run
        - kill: Stop a model_run that is currently running
        - remove: Remove a model_run that is waiting to be executed
        - resume: Warm start a model_run
        """
        data_interface = current_app.config.data_interface

        try:
            if action is None:
                data = request.get_json() or request.form
                data_interface.write_model_run(data)
            elif action == 'start':
                data = request.get_json() or request.form
                args = {
                    'verbosity': data['args']['verbosity'],
                    'warm_start': data['args']['warm_start'],
                    'output_format': data['args']['output_format']
                }
                if hasattr(data_interface, 'model_base_folder'):
                    args['directory'] = data_interface.model_base_folder
                current_app.config.scheduler.add(model_run_name, args)
            elif action == 'kill':
                current_app.config.scheduler.kill(model_run_name)
            elif action == 'remove':
                raise NotImplementedError
            elif action == 'resume':
                raise NotImplementedError
            else:
                raise SyntaxError("ModelRun action '%s' does not exist" % action)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 201
        return response

    def put(self, model_run_name):
        """Update a model_run:
        PUT /api/v1/model_runs
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data_interface.update_model_run(model_run_name, data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 200
        return response

    def delete(self, model_run_name):
        """Delete a model_run:
        DELETE /api/v1/model_runs
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_model_run(model_run_name)
        response = jsonify({})
        return response


class SosModelAPI(MethodView):
    """Implement CRUD operations for sos_model configuration data
    """
    def get(self, sos_model_name):
        """Get sos_model
        all: GET /api/v1/sos_models/
        one: GET /api/vi/sos_models/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface

        try:
            if sos_model_name is None:
                data = []
                data = data_interface.read_sos_models()
            else:
                data = {}
                data = data_interface.read_sos_model(sos_model_name)

            response = jsonify({
                'data': data,
                'error': {}
            })
        except SmifException as err:
            response = jsonify({
                'data': data,
                'error': parse_exceptions(err)
            })

        return response

    def post(self):
        """Create a sos_model:
        POST /api/v1/sos_models
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data_interface.write_sos_model(data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 201
        return response

    def put(self, sos_model_name):
        """Update a sos_model:
        PUT /api/v1/sos_models
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data_interface.update_sos_model(sos_model_name, data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 200
        return response

    def delete(self, sos_model_name):
        """Delete a sos_model:
        DELETE /api/v1/sos_models
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_sos_model(sos_model_name)
        response = jsonify({})
        return response


class SectorModelAPI(MethodView):
    """Implement CRUD operations for sector_model configuration data
    """
    def get(self, sector_model_name):
        """Get sector_models
        all: GET /api/v1/sector_models/
        one: GET /api/vi/sector_models/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface

        try:
            if sector_model_name is None:
                data = []
                data = data_interface.read_models(skip_coords=True)
            else:
                data = {}
                data = data_interface.read_model(sector_model_name, skip_coords=True)

            response = jsonify({
                'data': data,
                'error': {}
            })
        except SmifException as err:
            response = jsonify({
                'data': data,
                'error': parse_exceptions(err)
            })
        return response

    def post(self):
        """Create a sector_model:
        POST /api/v1/sector_models
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        try:
            data_interface.write_model(data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 201
        return response

    def put(self, sector_model_name):
        """Update a sector_model:
        PUT /api/v1/sector_models
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        try:
            data_interface.update_model(sector_model_name, data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 200
        return response

    def delete(self, sector_model_name):
        """Delete a sector_model:
        DELETE /api/v1/sector_models
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_model(sector_model_name)
        response = jsonify({})
        return response


class ScenarioAPI(MethodView):
    """Implement CRUD operations for scenarios configuration data
    """
    def get(self, scenario_name):
        """Get scenarios
        all: GET /api/v1/scenarios/
        one: GET /api/vi/scenarios/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface

        try:
            if scenario_name is None:
                data = []
                data = data_interface.read_scenarios(skip_coords=True)
            else:
                data = {}
                data = data_interface.read_scenario(scenario_name, skip_coords=True)

            response = jsonify({
                'data': data,
                'error': {}
            })
        except SmifException as err:
            response = jsonify({
                'data': data,
                'error': parse_exceptions(err)
            })

        return response

    def post(self):
        """Create a scenario:
        POST /api/v1/scenarios
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data = check_timestamp(data)
            data_interface.write_scenario(data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 201
        return response

    def put(self, scenario_name):
        """Update a scenario:
        PUT /api/v1/scenarios
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data = check_timestamp(data)
            data_interface.update_scenario(scenario_name, data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 200
        return response

    def delete(self, scenario_name):
        """Delete a scenario:
        DELETE /api/v1/scenarios
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_scenario(scenario_name)
        response = jsonify({})
        return response


class DimensionAPI(MethodView):
    """Implement CRUD operations for dimensions configuration data
    """
    def get(self, dimension_name):
        """Get dimensions
        all: GET /api/v1/dimensions/
        one: GET /api/vi/dimensions/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface

        try:
            if dimension_name is None:
                data = []
                data = data_interface.read_dimensions(skip_coords=True)
            else:
                data = {}
                data = data_interface.read_dimension(dimension_name, skip_coords=True)

            response = jsonify({
                'data': data,
                'error': {}
            })
        except SmifException as err:
            response = jsonify({
                'data': data,
                'error': parse_exceptions(err)
            })

        return response

    def post(self):
        """Create a dimension:
        POST /api/v1/dimensions
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data = check_timestamp(data)
            data_interface.write_dimension(data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 201
        return response

    def put(self, dimension_name):
        """Update a dimension:
        PUT /api/v1/dimensions
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        try:
            data = check_timestamp(data)
            data_interface.update_dimension(dimension_name, data)
        except SmifException as err:
            response = jsonify({
                'message': 'failed',
                'data': data,
                'error': parse_exceptions(err)
            })
        else:
            response = jsonify({"message": "success"})

        response.status_code = 200
        return response

    def delete(self, dimension_name):
        """Delete a dimension:
        DELETE /api/v1/dimensions
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_dimension(dimension_name)
        response = jsonify({})
        return response


def check_timestamp(data):
    """Check for timestamp and parse to datetime object
    """
    if 'stamp' in data:
        try:
            data['stamp'] = dateutil.parser.parse(data['stamp'])
        except(ValueError):
            pass
    return data


def parse_exceptions(exception):
    """Parse a group of exceptions so that it can be sent over
    the http-api
    """
    if type(exception) == SmifDataError:
        msg = defaultdict(list)
        for ex in exception.args[0]:
            msg[str(type(ex).__name__)].append(_parse_exception(ex))
    else:
        msg = {}
        msg[str(type(exception).__name__)] = [_parse_exception(exception)]

    return msg


def _parse_exception(ex):
    """Parse a single exception so that it can be sent over
    the http-api
    """
    if type(ex) == SmifValidationError:
        msg = ex.args[0]
    if type(ex) == SmifDataInputError:
        msg = {
            'component': ex.component,
            'error': ex.error,
            'message': ex.message,
        }
    if type(ex) == SmifDataNotFoundError:
        msg = ex.args[0]
    return msg



"""Dependencies represent the flow of data explicitly, with reference to the source model
output and sink model input.

A Dependency is an edge in the model process graph. No processing happens on dependency edges,
only on the Model nodes.

A Dependency can have a timestep specified - by default this is the current timestep, but a
system-of-systems model may also want to specify that data should be provided from the previous
timestep or a base timestep.
"""
from smif.metadata import RelativeTimestep


class Dependency(object):
    """Link a model input to a data source

    Arguments
    ---------
    source_model : smif.model.Model
        The source model object
    source : smif.metadata.Spec
        The source output description
    sink_model : smif.model.Model
        The sink model object
    sink : smif.metadata.Spec
        The sink input description
    timestep : smif.metadata.RelativeTimestep, optional
        The relative timestep of the dependency, defaults to CURRENT
    """
    def __init__(self, source_model, source, sink_model, sink,
                 timestep=RelativeTimestep.CURRENT):
        # Insist on identical metadata - conversions must be explicit
        if source != sink:
            diff = ""
            if source.dtype != sink.dtype:
                diff += "dtype(%s!=%s) " % (source.dtype, sink.dtype)
            if source.dims != sink.dims:
                diff += "dims(%s!=%s) " % (source.dims, sink.dims)
            if source.coords != sink.coords:
                diff += "coords do not match "
            if source.unit != sink.unit:
                diff += "unit(%s!=%s) " % (source.unit, sink.unit)
            msg = "Dependencies must connect identical metadata (up to variable name). " + \
                "Connecting %s:%s->%s:%s with mismatched %s"
            raise ValueError(
                msg % (source_model.name, source.name, sink_model.name, sink.name, diff))

        self.source_model = source_model
        self.source = source
        self.sink_model = sink_model
        self.sink = sink
        self.timestep = timestep

    def as_dict(self):
        """Serialise to dictionary representation
        """
        config = {
            'source': self.source_model.name,
            'source_output': self.source.name,
            'sink': self.sink_model.name,
            'sink_input': self.sink.name
        }
        try:
            config['timestep'] = self.timestep.value
        except AttributeError:
            config['timestep'] = self.timestep
        return config

    def __repr__(self):
        return "<Dependency({}, {}, {}, {}, {})>".format(
            self.source_model, self.source, self.sink_model, self.sink, self.timestep)

    def __eq__(self, other):
        return self.source_model == other.source_model \
            and self.source == other.source \
            and self.sink_model == other.sink_model \
            and self.sink == other.sink \
            and self.timestep == other.timestep



import datetime
import logging
import logging.config
import re
import sys
from collections import OrderedDict

LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s %(name)-12s: %(levelname)-8s %(message)s'
        },
        'message': {
            'format': '\033[1;34m%(levelname)-8s\033[0m %(message)s'
        }
    },
    'handlers': {
        'file': {
            'class': 'logging.FileHandler',
            'level': 'DEBUG',
            'formatter': 'default',
            'filename': 'smif.log',
            'mode': 'a',
            'encoding': 'utf-8'
        },
        'stream': {
            'class': 'logging.StreamHandler',
            'formatter': 'message',
            'level': 'DEBUG'
        }
    },
    'root': {
        'handlers': ['file', 'stream'],
        'level': 'DEBUG'
    }
}


# Make profiling methods available through the logger
def profiling_start(self, operation, key):
    time_placeholder = datetime.time(0, 0)
    level = sum(log[1]['stop'] == time_placeholder for log in logging.Logger._profile.items())
    logging.Logger._profile[(operation, key)] = {
        'start': datetime.datetime.now(),
        'stop': time_placeholder,
        'level': level
    }


def profiling_stop(self, operation, key):
    logging.Logger._profile[(operation, key)]['stop'] = datetime.datetime.now()


def summary(self, *args, **kws):

    if self.isEnabledFor(logging.INFO):
        summary = []
        columns = [30, 80, 12]
        column = "{:" + str(columns[0]) + "s}" + \
                 "{:" + str(columns[1]) + "s}" + \
                 "{:" + str(columns[2]) + "s}"

        level_width = max([profile['level'] for profile
                          in logging.Logger._profile.values()]) * 2
        total_width = sum(columns) + level_width

        # header
        summary.append(("{:*^" + str(total_width) + "s}").format(" Modelrun time profile "))
        summary.append(column.format('Function', 'Operation', 'Duration [hh:mm:ss]'))
        summary.append("*"*total_width)

        # body
        for profile in logging.Logger._profile.keys():

            # calculate time diff
            profile_data = logging.Logger._profile[profile]
            diff = profile_data['stop'] - profile_data['start']
            s = diff.total_seconds()
            time_spent = '{:02d}:{:02d}:{:02d}'.format(
                int(s // 3600), int(s % 3600 // 60), int(s % 60))

            # trunctuate long lines
            if len(profile[0]) > columns[0]-2:
                func = profile[0][:columns[0]-3] + '..'
            else:
                func = profile[0]
            if len(profile[1]) > columns[1]-2:
                op = profile[1][:columns[1]-3] + '..'
            else:
                op = profile[1]

            summary.append(profile_data['level']*'| ' + column.format(
                func, op, time_spent))

        # footer
        summary.append("*"*total_width)

        for entry in summary:
            self._log(logging.INFO, entry, args)


logging.Logger.profiling_start = profiling_start
logging.Logger.profiling_stop = profiling_stop
logging.Logger.summary = summary
logging.Logger._profile = OrderedDict()

# Configure logging once, outside of any dependency on argparse
VERBOSITY = None
if '--verbose' in sys.argv:
    VERBOSITY = sys.argv.count('--verbose')
else:
    for arg in sys.argv:
        if re.match(r'\A-v+\Z', arg):
            VERBOSITY = len(arg) - 1
            break

if VERBOSITY is None:
    LOGGING_CONFIG['root']['level'] = logging.WARNING
elif VERBOSITY == 1:
    LOGGING_CONFIG['root']['level'] = logging.INFO
else:
    LOGGING_CONFIG['root']['level'] = logging.DEBUG

logging.config.dictConfig(LOGGING_CONFIG)
LOGGER = logging.getLogger(__name__)
LOGGER.debug('Debug logging enabled.')



from unittest.mock import Mock, PropertyMock

from pytest import fixture, raises
from smif.decision.decision import DecisionManager, PreSpecified, RuleBased
from smif.exception import SmifDataNotFoundError


@fixture(scope='function')
def plan():
    planned_interventions = [
        {'name': 'small_pumping_station_oxford', 'build_year': 2010},
        {'name': 'small_pumping_station_abingdon', 'build_year': 2015},
        {'name': 'large_pumping_station_oxford', 'build_year': 2020}
    ]

    return planned_interventions


@fixture(scope='function')
def get_strategies():
    strategies = [{'strategy': 'pre-specified-planning',
                   'description': 'build_nuclear',
                   'interventions': [
                       {'name': 'nuclear_large', 'build_year': 2012},
                       {'name': 'carrington_retire', 'build_year': 2011}]
                   }]

    return strategies


@fixture(scope='function')
def get_register():
    lifetime = {'technical_lifetime': {'value': 99}}
    register = {'nuclear_large': lifetime,
                'carrington_retire': lifetime,
                'small_pumping_station_oxford': lifetime,
                'small_pumping_station_abingdon': lifetime,
                'large_pumping_station_oxford': lifetime
                }
    return register


class TestPreSpecified:

    def test_initialisation(self, plan):

        timesteps = [2010, 2015, 2020]

        actual = PreSpecified(timesteps, Mock(), plan)

        assert actual.timesteps == timesteps

    def test_generator(self, plan):

        timesteps = [2010, 2015, 2020]
        dm = PreSpecified(timesteps, Mock(), plan)

        actual = next(dm)

        expected = {
            'decision_iterations': [0],
            'timesteps': timesteps
        }

        assert actual == expected

    def test_get_decision(self, plan, get_register):

        register = get_register

        timesteps = [2010, 2015, 2020]
        dm = PreSpecified(timesteps, register, plan)

        mock_handle = Mock()
        type(mock_handle).current_timestep = PropertyMock(return_value=2010)
        actual = dm.get_decision(mock_handle)
        expected = [
            {'name': 'small_pumping_station_oxford',
             'build_year': 2010}]
        assert actual == expected

        type(mock_handle).current_timestep = PropertyMock(return_value=2015)
        actual = dm.get_decision(mock_handle)
        expected = [
            {'name': 'small_pumping_station_oxford',
             'build_year': 2010},
            {'name': 'small_pumping_station_abingdon',
             'build_year': 2015}]
        assert actual == expected

        type(mock_handle).current_timestep = PropertyMock(return_value=2020)
        actual = dm.get_decision(mock_handle)
        expected = [
            {'name': 'small_pumping_station_oxford',
             'build_year': 2010},
            {'name': 'small_pumping_station_abingdon',
             'build_year': 2015},
            {'name': 'large_pumping_station_oxford',
             'build_year': 2020}
        ]
        assert actual == expected

    def test_get_decision_two(self, get_strategies, get_register):
        register = get_register
        dm = PreSpecified([2010, 2015], register, get_strategies[0]['interventions'])

        mock_handle = Mock()
        type(mock_handle).current_timestep = PropertyMock(return_value=2010)
        actual = dm.get_decision(mock_handle)
        expected = [
            {'name': 'nuclear_large', 'build_year': 2012},
            {'name': 'carrington_retire', 'build_year': 2011}
        ]
        # assert actual == expected
        # we don't mind the order
        assert (actual) == (expected)

        # actual = dm.get_decision(2015)
        # expected = [('carrington_retire', 2011)]
        # assert actual == expected

        type(mock_handle).current_timestep = PropertyMock(return_value=2015)
        actual = dm.get_decision(mock_handle)
        expected = [
            {'name': 'nuclear_large', 'build_year': 2012},
            {'name': 'carrington_retire', 'build_year': 2011}
        ]
        assert (actual) == (expected)

    def test_buildable(self, get_strategies):
        dm = PreSpecified([2010, 2015], Mock(), get_strategies[0]['interventions'])
        assert dm.timesteps == [2010, 2015]
        assert dm.buildable(2010, 2010) is True
        assert dm.buildable(2011, 2010) is True

    def test_historical_intervention_buildable(self, get_strategies):
        dm = PreSpecified([2020, 2030], Mock(), get_strategies[0]['interventions'])
        assert dm.timesteps == [2020, 2030]
        assert dm.buildable(1980, 2020) is True
        assert dm.buildable(1990, 2020) is True

    def test_buildable_raises(self, get_strategies):
        dm = PreSpecified([2010, 2015], Mock(), get_strategies[0]['interventions'])
        with raises(ValueError):
            dm.buildable(2015, 2014)

    def test_within_lifetime(self):
        dm = PreSpecified([2010, 2015], Mock(), [])
        assert dm.within_lifetime(2010, 2010, 1)

    def test_within_lifetime_does_not_check_start(self):
        """Note that the ``within_lifetime`` method does not check
        that the build year is compatible with timestep
        """
        dm = PreSpecified([2010, 2015], Mock(), [])
        assert dm.within_lifetime(2011, 2010, 1)

    def test_negative_lifetime_raises(self):
        dm = PreSpecified([2010, 2015], Mock(), [])
        with raises(ValueError):
            dm.within_lifetime(2010, 2010, -1)


class TestRuleBasedProperties:

    def test_timesteps(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        assert dm.current_timestep == 2010
        assert dm.next_timestep == 2015

    def test_timesteps_end(self):
        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        dm.current_timestep = 2020
        assert dm.next_timestep is None

    def test_timesteps_begin(self):
        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        dm.current_timestep = 2010
        assert dm.previous_timestep is None

    def test_timesteps_first_last(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        assert dm.first_timestep == 2010
        assert dm.last_timestep == 2020

    def test_interventions(self):

        interventions = Mock()

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, interventions)
        assert dm.interventions == interventions

    def test_get_intervention(self):

        interventions = {'a': {'name': 'a'},
                         'b': {'name': 'b'},
                         'c': {'name': 'c'}}

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, interventions)
        assert dm.get_intervention('a') == interventions['a']
        with raises(SmifDataNotFoundError) as ex:
            dm.get_intervention('z')
        msg = "Intervention 'z' is not found in the list of available interventions"
        assert msg in str(ex)


class TestRuleBased:

    def test_initialisation(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        assert dm.timesteps == timesteps
        assert dm.satisfied is False
        assert dm.current_timestep == 2010
        assert dm.current_iteration == 0

    def test_generator(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())

        actual = next(dm)
        assert actual == {
            'decision_iterations': [1],
            'timesteps': [2010]
        }

        dm.satisfied = True
        actual = next(dm)
        assert actual == {
            'decision_iterations': [2],
            'timesteps': [2015],
            'decision_links': {
                2: 1
            }
        }
        assert dm.satisfied is False

        actual = next(dm)
        assert actual == {
            'decision_iterations': [3],
            'timesteps': [2015],
            'decision_links': {
                3: 1
            }
        }
        assert dm.satisfied is False

        dm.satisfied = True
        dm.current_timestep = 2020
        actual = next(dm)
        assert actual is None


class TestDecisionManager():

    @fixture(scope='function')
    def decision_manager(self, empty_store):
        empty_store.write_model_run({'name': 'test', 'sos_model': 'test_sos_model'})
        empty_store.write_sos_model({'name': 'test_sos_model', 'sector_models': []})
        empty_store.write_strategies('test', [])
        sos_model = Mock()
        sos_model.name = 'test_sos_model'
        sos_model.sector_models = []

        df = DecisionManager(empty_store, [2010, 2015], 'test', sos_model)
        return df

    def test_decision_manager_init(self, decision_manager):
        df = decision_manager
        dm = df.decision_loop()
        bundle = next(dm)
        assert bundle == {
            'decision_iterations': [0],
            'timesteps': [2010, 2015]
        }
        with raises(StopIteration):
            next(dm)

    def test_available_interventions(self, decision_manager):
        df = decision_manager
        df._register = {'a': {'name': 'a'},
                        'b': {'name': 'b'},
                        'c': {'name': 'c'}}

        assert df.available_interventions == df._register

        df.planned_interventions = {'a', 'b'}

        expected = {'c': {'name': 'c'}}

        assert df.available_interventions == expected

    def test_get_intervention(self, decision_manager):
        df = decision_manager
        df._register = {'a': {'name': 'a'},
                        'b': {'name': 'b'},
                        'c': {'name': 'c'}}

        assert df.get_intervention('a') == {'name': 'a'}

        with raises(SmifDataNotFoundError):
            df.get_intervention('z')



"""Test all MetadataStore implementations
"""
from pytest import fixture, mark, param
from smif.data_layer.database_interface import DbMetadataStore
from smif.data_layer.file.file_metadata_store import FileMetadataStore
from smif.data_layer.memory_interface import MemoryMetadataStore


@fixture(
    params=[
        'memory',
        'file',
        param('database', marks=mark.skip)
    ])
def init_handler(request, setup_empty_folder_structure):
    if request.param == 'memory':
        handler = MemoryMetadataStore()
    elif request.param == 'file':
        base_folder = setup_empty_folder_structure
        handler = FileMetadataStore(base_folder)
    elif request.param == 'database':
        handler = DbMetadataStore()
        raise NotImplementedError

    return handler


@fixture
def handler(init_handler, unit_definitions, dimension, sample_dimensions):
    handler = init_handler

    # metadata
    handler.write_unit_definitions(unit_definitions)
    handler.write_dimension(dimension)
    for dim in sample_dimensions:
        handler.write_dimension(dim)

    return handler


class TestUnits():
    """Read units definitions
    """
    def test_read_units(self, handler, unit_definitions):
        expected = unit_definitions
        actual = handler.read_unit_definitions()
        assert actual == expected


class TestDimensions():
    """Read/write/update/delete dimensions
    """
    def test_read_dimensions(self, handler, dimension, sample_dimensions):
        actual = handler.read_dimensions()
        expected = [dimension] + sample_dimensions
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_read_dimension(self, handler, dimension):
        assert handler.read_dimension('category') == dimension

    def test_write_dimension(self, handler, dimension, sample_dimensions):
        another_dimension = {'name': '3rd', 'elements': [{'name': 'a'}, {'name': 'b'}]}
        handler.write_dimension(another_dimension)
        actual = handler.read_dimensions()
        expected = [dimension, another_dimension] + sample_dimensions
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_write_interval_dimension(self, handler):
        intervals = {
            'name': 'seasons',
            'elements': [
                {'name': 'spring', 'interval': [['P2M', 'P5M']]},
                {'name': 'summer', 'interval': [['P5M', 'P8M']]},
                {'name': 'autumn', 'interval': [['P8M', 'P11M']]},
                {'name': 'winter', 'interval': [['P0M', 'P2M'], ['P11M', 'P12M']]},
            ]
        }
        handler.write_dimension(intervals)
        actual = handler.read_dimension('seasons')
        assert actual == intervals

    def test_update_dimension(self, handler, dimension, sample_dimensions):
        another_dimension = {'name': 'category', 'elements': [{'name': 4}, {'name': 5}, {'name': 6}]}
        handler.update_dimension('category', another_dimension)
        actual = handler.read_dimensions()
        expected = [another_dimension] + sample_dimensions
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_delete_dimension(self, handler, sample_dimensions):
        handler.delete_dimension('category')
        actual = handler.read_dimensions()
        expected = sample_dimensions
        assert sorted_by_name(actual) == sorted_by_name(expected)


def sorted_by_name(list_):
    """Helper to sort lists-of-dicts
    """
    return sorted(list_, key=lambda d: d['name'])



"""Provide an interface to data, parameters and results

A :class:`DataHandle` is passed in to a :class:`Model` at runtime, to provide
transparent access to the relevant data and parameters for the current
:class:`ModelRun` and iteration. It gives read access to parameters and input
data (at any computed or pre-computed timestep) and write access to output data
(at the current timestep).
"""
from copy import copy
from logging import getLogger
from types import MappingProxyType
from typing import Dict, List

import numpy as np  # type: ignore
from smif.data_layer.store import Store
from smif.data_layer.data_array import DataArray
from smif.exception import SmifDataError
from smif.metadata import RelativeTimestep


class DataHandle(object):
    """Get/set model parameters and data
    """
    def __init__(self, store: Store, modelrun_name, current_timestep, timesteps, model,
                 decision_iteration=None):
        """Create a DataHandle for a Model to access data, parameters and state, and to
        communicate results.

        Parameters
        ----------
        store : Store
            Backing store for inputs, parameters, results
        modelrun_name : str
            Name of the current modelrun
        model : Model
            Model which will use this DataHandle
        decision_iteration : int, default=None
            ID of the current Decision iteration
        state : list, default=None
        """
        self.logger = getLogger(__name__)
        self._store = store
        self._modelrun_name = modelrun_name
        self._current_timestep = current_timestep
        self._timesteps = timesteps
        self._decision_iteration = decision_iteration

        self._model_name = model.name
        self._inputs = model.inputs
        self._outputs = model.outputs
        self._model = model

        modelrun = self._store.read_model_run(self._modelrun_name)
        sos_model = self._store.read_sos_model(modelrun['sos_model'])

        self._scenario_dependencies = {}  # type: Dict[str, Dict]
        self._model_dependencies = {}  # type: Dict[str, Dict]
        scenario_variants = modelrun['scenarios']
        self._load_dependencies(sos_model, scenario_variants)
        self.logger.debug(
            "Create with %s model, %s scenario dependencies",
            len(self._scenario_dependencies),
            len(self._model_dependencies))

        self._parameters = {}  # type: Dict[str, DataArray]
        self._load_parameters(sos_model, modelrun['narratives'])

    def _load_dependencies(self, sos_model, scenario_variants):
        """Load Model dependencies as a dict with {input_name: list[Dependency]}
        """
        for dep in sos_model['model_dependencies']:
            if dep['sink'] == self._model_name:
                input_name = dep['sink_input']
                self._model_dependencies[input_name] = {
                    'source_model_name': dep['source'],
                    'source_output_name': dep['source_output'],
                    'type': 'model'
                }

        for dep in sos_model['scenario_dependencies']:
            if dep['sink'] == self._model_name:
                input_name = dep['sink_input']
                self._scenario_dependencies[input_name] = {
                    'source_model_name': dep['source'],
                    'source_output_name': dep['source_output'],
                    'type': 'scenario',
                    'variant': scenario_variants[dep['source']]
                }

    def _load_parameters(self, sos_model, concrete_narratives):
        """Load parameter values for model run

        Parameters for each of the contained sector models are loaded
        into memory as a data_handle is initialised.

        Firstly, default values for the parameters are loaded from the parameter
        specs contained within each of the sector models

        Then, the data from the list of narrative variants linked to the current
        model run are loaded into the parameters contained within the

        Arguments
        ---------
        sos_model : dict
            A configuration dictionary of a system-of-systems model
        concrete_narratives: dict of list
            Links narrative names to a list of variants to furnish parameters
            with values {narrative_name: [variant_name, ...]}
        """
        # Populate the parameters with their default values
        for parameter in self._model.parameters.values():
            self._parameters[parameter.name] = \
                self._store.read_model_parameter_default(self._model.name, parameter.name)

        # Load in the concrete narrative and selected variants from the model run
        for narrative_name, variant_names in concrete_narratives.items():
            # Load the narrative
            try:
                narrative = [x for x in sos_model['narratives']
                             if x['name'] == narrative_name][0]
            except IndexError:
                msg = "Couldn't find a match for {} in {}"
                raise IndexError(msg.format(narrative_name, sos_model['name']))
            self.logger.debug("Loaded narrative: %s", narrative)
            self.logger.debug("Considering variants: %s", variant_names)

            # Read parameter data from each variant, later variants overriding
            # previous parameter values
            for variant_name in variant_names:
                try:
                    parameter_list = narrative['provides'][self._model.name]
                except KeyError:
                    parameter_list = []

                for parameter in parameter_list:
                    da = self._store.read_narrative_variant_data(
                        sos_model['name'],
                        narrative_name, variant_name, parameter
                    )
                    self._parameters[parameter].update(da)

    def derive_for(self, model):
        """Derive a new DataHandle configured for the given Model

        Parameters
        ----------
        model : Model
            Model which will use this DataHandle
        """
        return DataHandle(
            store=self._store,
            modelrun_name=self._modelrun_name,
            current_timestep=self._current_timestep,
            timesteps=list(self.timesteps),
            model=model,
            decision_iteration=self._decision_iteration
        )

    def __getitem__(self, key):
        if key in self._parameters:
            return self.get_parameter(key)
        elif key in self._inputs:
            return self.get_data(key)
        elif key in self._outputs:
            return self.get_results(key)
        else:
            raise KeyError(
                "'%s' not recognised as input, output or parameter for '%s'" %
                (key, self._model_name)
                )

    def __setitem__(self, key, value):
        if hasattr(value, 'as_ndarray'):
            raise TypeError("Pass in a numpy array")
        self.set_results(key, value)

    @property
    def current_timestep(self):
        """Current timestep
        """
        return self._current_timestep

    @property
    def previous_timestep(self):
        """Previous timestep
        """
        return RelativeTimestep.PREVIOUS.resolve_relative_to(
            self._current_timestep,
            self._timesteps
        )

    @property
    def base_timestep(self):
        """Base timestep
        """
        return RelativeTimestep.BASE.resolve_relative_to(
            self._current_timestep,
            self._timesteps
        )

    @property
    def timesteps(self):
        """All timesteps (as tuple)
        """
        return tuple(self._timesteps)

    @property
    def decision_iteration(self):
        return self._decision_iteration

    def get_state(self):
        """The current state of the model

        If the DataHandle instance has a timestep, then state is
        established from the state file.

        Returns
        -------
        list of tuple
            A list of (intervention name, build_year) installed in the current timestep

        Raises
        ------
        ValueError
            If self._current_timestep is None an error is raised.
        """
        if self._current_timestep is None:
            raise ValueError("You must pass a timestep value to get state")
        else:

            sos_state = self._store.read_state(
                self._modelrun_name,
                self._current_timestep,
                self._decision_iteration
            )

        return sos_state

    def get_current_interventions(self):
        """Get the interventions that exist in the current state

        Returns
        -------
        dict of dicts
            A dict of intervention dicts with build_year attribute keyed by name
        """
        state = self.get_state()

        current_interventions = {}
        all_interventions = self._store.read_interventions(self._model_name)

        for decision in state:
            name = decision['name']
            build_year = decision['build_year']
            try:
                serialised = all_interventions[name]
                serialised['build_year'] = build_year
                current_interventions[name] = serialised
            except KeyError:
                # ignore if intervention is not in current set
                pass

        msg = "State matched with %s interventions"
        self.logger.info(msg, len(current_interventions))

        return current_interventions

    def get_data(self, input_name: str, timestep=None) -> DataArray:
        """Get data required for model inputs

        Parameters
        ----------
        input_name : str
        timestep : RelativeTimestep or int, optional
            defaults to RelativeTimestep.CURRENT

        Returns
        -------
        smif.data_layer.data_array.DataArray
            Contains data annotated with the metadata and provides utility methods
            to access the data in different ways

        Raises
        ------
        SmifDataError
            If any data reading error occurs below this method, the error is
            handled and reraised within the context of the current call
        """
        if input_name not in self._inputs:
            raise KeyError(
                "'{}' not recognised as input for '{}'".format(input_name, self._model_name))

        timestep = self._resolve_timestep(timestep)

        dep = self._resolve_source(input_name)

        self.logger.debug(
            "Read %s %s %s", dep['source_model_name'], dep['source_output_name'],
            timestep)

        if dep['type'] == 'scenario':
            data = self._get_scenario(dep, timestep, input_name)
        else:
            input_spec = self._inputs[input_name]
            data = self._get_result(dep, timestep, input_spec)

        return data

    def _resolve_timestep(self, timestep):
        """Resolves a relative timestep to an absolute timestep

        Arguments
        ---------
        timestep : RelativeTimestep or int

        Returns
        -------
        int
        """
        if self._current_timestep is None:
            if timestep is None:
                raise ValueError("You must provide a timestep to obtain data")
            elif hasattr(timestep, "resolve_relative_to"):
                timestep = timestep.resolve_relative_to(self._timesteps[0], self._timesteps)
            else:
                assert isinstance(timestep, int) and timestep in self._timesteps
        else:
            if timestep is None:
                timestep = self._current_timestep
            elif hasattr(timestep, "resolve_relative_to"):
                timestep = timestep.resolve_relative_to(self._current_timestep,
                                                        self._timesteps)
            else:
                assert isinstance(timestep, int) and timestep <= self._current_timestep
        return timestep

    def _get_result(self, dep, timestep, input_spec) -> DataArray:
        """Retrieves a model result for a dependency
        """
        output_spec = copy(input_spec)
        output_spec.name = dep['source_output_name']
        self.logger.debug("Getting model result for %s via %s from %s", input_spec, dep, output_spec)
        try:
            data = self._store.read_results(
                self._modelrun_name,
                dep['source_model_name'],  # read from source model
                output_spec,  # using source model output spec
                timestep,
                self._decision_iteration
            )
            data.name = input_spec.name  # ensure name matches input (as caller expects)
        except SmifDataError as ex:
            msg = "Could not read data for output '{}' from '{}' in {}, iteration {}"
            raise SmifDataError(msg.format(
                output_spec.name,
                dep['source_model_name'],
                timestep,
                self._decision_iteration
            )) from ex
        return data

    def _get_scenario(self, dep, timestep, input_name) -> DataArray:
        """Retrieves data from a scenario

        Arguments
        ---------
        dep : dict
            A scenario dependency
        timestep : int

        Returns
        -------
        DataArray
        """
        try:
            data = self._store.read_scenario_variant_data(
                dep['source_model_name'],  # read from a given scenario model
                dep['variant'],  # with given scenario variant
                dep['source_output_name'],  # using output (variable) name
                timestep
            )
            data.name = input_name  # ensure name matches input (as caller expects)
        except SmifDataError as ex:
            msg = "Could not read data for output '{}' from '{}.{}' in {}"
            raise SmifDataError(msg.format(
                dep['source_output_name'],
                dep['source_model_name'],
                dep['variant'],
                timestep
            )) from ex
        return data

    def _resolve_source(self, input_name) -> Dict:
        """Find best dependency to provide input data

        Returns
        -------
        dep : dict
            A scenario or model dependency dictionary
        """
        scenario_dep = None
        try:
            scenario_dep = self._scenario_dependencies[input_name]
        except KeyError:
            pass

        model_dep = None
        try:
            model_dep = self._model_dependencies[input_name]
        except KeyError:
            pass

        if scenario_dep is not None and model_dep is not None:
            # if multiple dependencies, use scenario for timestep 0, model for
            # subsequent timesteps
            if self._current_timestep == self._timesteps[0]:
                dep = scenario_dep
            else:
                dep = model_dep
        elif scenario_dep is not None:
            # else assume single dependency per input
            dep = scenario_dep
        elif model_dep is not None:
            dep = model_dep
        else:
            raise SmifDataError("Dependency not defined for input '{}' in model '{}'".format(
                input_name, self._model_name
            ))
        return dep

    def get_base_timestep_data(self, input_name):
        """Get data from the base timestep as required for model inputs

        Parameters
        ----------
        input_name : str

        Returns
        -------
        smif.data_layer.data_array.DataArray
        """
        return self.get_data(input_name, RelativeTimestep.BASE)

    def get_previous_timestep_data(self, input_name):
        """Get data from the previous timestep as required for model inputs

        Parameters
        ----------
        input_name : str

        Returns
        -------
        smif.data_layer.data_array.DataArray
        """
        return self.get_data(input_name, RelativeTimestep.PREVIOUS)

    def get_parameter(self, parameter_name):
        """Get the value for a  parameter

        Parameters
        ----------
        parameter_name : str

        Returns
        -------
        smif.data_layer.data_array.DataArray
            Contains data annotated with the metadata and provides utility methods
            to access the data in different ways
        """
        if parameter_name not in self._parameters:
            raise KeyError(
                "'{}' not recognised as parameter for '{}'".format(
                    parameter_name, self._model_name))

        return self._parameters[parameter_name]

    def get_parameters(self):
        """Get all parameter values

        Returns
        -------
        parameters : MappingProxyType
            Read-only view of parameters (like a read-only dict)
        """
        return MappingProxyType(self._parameters)

    def set_results(self, output_name, data):
        """Set results values for model outputs

        Parameters
        ----------
        output_name : str
        data : numpy.ndarray
        """
        if hasattr(data, 'as_ndarray'):
            raise TypeError("Pass in a numpy array")

        if output_name not in self._outputs:
            raise KeyError(
                "'{}' not recognised as output for '{}'".format(output_name, self._model_name))

        self.logger.debug(
            "Write %s %s %s", self._model_name, output_name, self._current_timestep)

        spec = self._outputs[output_name]

        da = DataArray(spec, data)

        self._store.write_results(
            da,
            self._modelrun_name,
            self._model_name,
            self._current_timestep,
            self._decision_iteration
        )

    def get_results(self, output_name, decision_iteration=None,
                    timestep=None):
        """Get results values for model outputs

        Parameters
        ----------
        output_name : str
            The name of an output for `model_name`
        decision_iteration : int, default=None
        timestep : int or RelativeTimestep, default=None

        Returns
        -------
        smif.data_layer.data_array.DataArray
            Contains data annotated with the metadata and provides utility methods
            to access the data in different ways

        Notes
        -----
        Access to model results is only granted to models contained
        within self._model if self._model is a smif.model.model.CompositeModel
        """
        model_name = self._model.name

        # resolve timestep
        if timestep is None:
            timestep = self._current_timestep
        elif isinstance(timestep, RelativeTimestep):
            timestep = timestep.resolve_relative_to(self._current_timestep, self._timesteps)
        else:
            assert isinstance(timestep, int) and timestep <= self._current_timestep

        # find output spec
        try:
            spec = self._model.outputs[output_name]
        except KeyError:
            msg = "'{}' not recognised as output for '{}'"
            raise KeyError(msg.format(output_name, model_name))

        if decision_iteration is None:
            decision_iteration = self._decision_iteration

        self.logger.debug(
            "Read %s %s %s", model_name, output_name, timestep)

        return self._store.read_results(
            self._modelrun_name,
            model_name,
            spec,
            timestep,
            decision_iteration
        )

    def read_unit_definitions(self) -> List[str]:
        """Read unit definitions

        Returns
        -------
        list[str]
        """
        return self._store.read_unit_definitions()

    def read_coefficients(self, source_dim: str, destination_dim: str) -> np.ndarray:
        """Reads coefficients from the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `write_coefficients` implement caching of conversion
        coefficients between dimensions.

        Parameters
        ----------
        source_dim: str
            Dimension name
        destination_dim: str
            Dimension name

        Returns
        -------
        numpy.ndarray
        """
        data = self._store.read_coefficients(source_dim, destination_dim)
        return data

    def write_coefficients(self, source_dim: str, destination_dim: str, data: np.ndarray):
        """Writes coefficients to the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `read_coefficients` implement caching of conversion
        coefficients between dimensions.

        Parameters
        ----------
        source_dim: str
            Dimension name
        destination_dim: str
            Dimension name
        data : numpy.ndarray
        """
        data = self._store.write_coefficients(source_dim, destination_dim, data)
        return data


class ResultsHandle(object):
    """Results access for decision modules
    """
    def __init__(self, store, modelrun_name, sos_model, current_timestep, timesteps=None,
                 decision_iteration=None):
        self._store = store
        self._modelrun_name = modelrun_name
        self._sos_model = sos_model

        self._current_timestep = current_timestep
        self._timesteps = timesteps
        self._decision_iteration = decision_iteration

    @property
    def base_timestep(self):
        return self._timesteps[0]

    @property
    def current_timestep(self):
        return self._current_timestep

    @property
    def previous_timestep(self):
        rel = RelativeTimestep.PREVIOUS
        return rel.resolve_relative_to(self._current_timestep, self._timesteps)

    @property
    def decision_iteration(self):
        return self._decision_iteration

    def get_results(self, model_name, output_name, timestep, decision_iteration):
        """Access model results

        Parameters
        ----------
        model_name : str
        output_name : str
        timestep : int
        decision_iteration : int

        Returns
        -------
        smif.data_layer.data_array.DataArray
            Contains data annotated with the metadata and provides utility methods
            to access the data in different ways
        """
        # resolve timestep
        if hasattr(timestep, 'resolve_relative_to'):
            timestep = timestep.resolve_relative_to(self._current_timestep, self._timesteps)
        else:
            assert isinstance(timestep, int) and timestep <= self._current_timestep

        if model_name in [model.name for model in self._sos_model.models]:
            results_model = self._sos_model.get_model(model_name)
        else:
            msg = "Model '{}' is not contained in SosModel '{}'. Found {}."
            raise KeyError(msg.format(model_name, self._sos_model.name,
                                      self._sos_model.models)
                           )

        try:
            spec = results_model.outputs[output_name]
        except KeyError:
            msg = "'{}' not recognised as output for '{}'"
            raise KeyError(msg.format(output_name, model_name))

        results = self._store.read_results(self._modelrun_name,
                                           model_name,
                                           spec,
                                           timestep,
                                           decision_iteration)

        return results



from copy import copy

from smif.decision.decision import RuleBased


class EnergyAgent(RuleBased):
    """A coupled power-producer/regulator decision algorithm for simulating
    decision making in an energy supply model

    EnergyAgent consists of two actors - a PowerProducer and a Regulator.

    The Regulator reviews the performance of the electricity system in the previous period
    against exogenously defined performance metric thresholds,
    such as emissions and capacity margin, and imposes constraints upon the
    available interventions through:
    - emission taxes
    - portfolio standards (e.g. % mandated to come from a generation type)

    In each timestep the PowerProducer assesses the current capacity shortfall
    against its projection of electricity demand and selects a portfolio of
    interventions using a heuristic based upon LCOE (levelised cost of electricity).

    """
    def __init__(self, timesteps, register):
        super().__init__(timesteps, register)
        self.model_name = 'energy_supply'

    @staticmethod
    def from_dict(config):
        timesteps = config['timesteps']
        register = config['register']
        return EnergyAgent(timesteps, register)

    def get_decision(self, data_handle):
        budget = self.run_regulator(data_handle)
        decisions = self.run_power_producer(data_handle, budget)
        return decisions

    def run_regulator(self, data_handle):
        """
        data_handle
        """
        budget = 100

        # TODO Should be the iteration previous to the current one?
        if data_handle.current_timestep > data_handle.base_timestep:
            previous_timestep = data_handle.previous_timestep
            iteration = self._max_iteration_by_timestep[previous_timestep]
            output_name = 'cost'
            cost = data_handle.get_results(model_name='energy_demand',
                                           output_name=output_name,
                                           decision_iteration=iteration,
                                           timestep=previous_timestep)
            budget -= sum(cost.as_ndarray())

        self.satisfied = True
        return budget

    def run_power_producer(self, data_handle, budget):
        """
        data_handle
        budget : float
        """
        cheapest_first = []
        for name, item in self.interventions.items():
            cheapest_first.append((name, float(item['capital_cost']['value'])))
        sorted(cheapest_first, key=lambda x: float(x[1]), reverse=True)

        within_budget = []
        remaining_budget = copy(budget)
        for intervention in cheapest_first:
            if intervention[1] <= remaining_budget:
                within_budget.append({'name': intervention[0],
                                      'build_year': data_handle.current_timestep})
                remaining_budget -= intervention[1]

        return within_budget



"""Schedulers are used to run models.

The defaults provided allow model runs to be scheduled as subprocesses,
or individual models to be called in series.

Future implementations may interface with common schedulers to queue
up models to run in parallel and/or distributed.
"""
import itertools
import logging
import subprocess
import traceback
from collections import defaultdict
from datetime import datetime

import networkx
from smif.data_layer import DataHandle
from smif.model import ModelOperation


class ModelRunScheduler(object):
    """The scheduler can run instances of smif as a subprocess
    and can provide information whether the modelrun is running,
    is done or has failed.
    """
    def __init__(self):
        self._status = defaultdict(lambda: 'unstarted')
        self._process = {}
        self._output = defaultdict(str)
        self._err = {}
        self.lock = False

    def add(self, model_run_name, args):
        """Add a model_run to the Modelrun scheduler.

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun
        args: dict
            Arguments for the command-line interface

        Exception
        ---------
        Exception
            When the modelrun was already started

        Notes
        -----
        There is no queuing mechanism implemented, each `add`
        will directly start a subprocess. This means that it
        is possible to run multiple modelruns concurrently.
        This may cause conflicts, it depends on the
        implementation whether a certain sector model / wrapper
        touches the filesystem or other shared resources.
        """
        if self._status[model_run_name] is not 'running':
            self._output[model_run_name] = ''
            self._status[model_run_name] = 'queing'

            smif_call = (
                'smif ' +
                '-'*(int(args['verbosity']) > 0) + 'v'*int(args['verbosity']) +
                ' run' + ' ' + model_run_name + ' ' +
                '-d' + ' ' + args['directory'] + ' ' +
                '-w'*args['warm_start'] + ' '*args['warm_start'] +
                '-i' + ' ' + args['output_format']
            )

            self._process[model_run_name] = subprocess.Popen(
                smif_call,
                shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            format_args = {
                'model_run_name': model_run_name,
                'datetime': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                'pid': str(self._process[model_run_name].pid),
                'smif_call': smif_call,
                'colour': "\x1b[1;34m",
                'reset': "\x1b[0m",
                'space': " \x1b"
            }
            format_str = """\
                {colour}Modelrun{reset} {model_run_name}
                {colour}Time{reset}     {datetime}
                {colour}PID{reset}      {pid}
                {colour}Command{reset}  {smif_call}
                """
            format_str.replace(" ", "{space}")
            output = format_str.format(**format_args)
            output += "-" * 100 + "\n"
            self._output[model_run_name] = output
            self._status[model_run_name] = 'running'
        else:
            raise Exception('Model is already running.')

    def kill(self, model_run_name):
        """ Kill a Modelrun that is already running

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun
        """
        if self._status[model_run_name] == 'running':
            self._process[model_run_name].kill()
            self._status[model_run_name] = 'stopped'

    def get_status(self, model_run_name):
        """Get the status from the Modelrun scheduler.

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun

        Returns
        -------
        dict: A message containing the status, command-line
        output and error that can be directly sent back over
        the http api.

        Notes
        -----
        Possible status:

        unstarted:
            Model run was not started
        queing:
            Model run is waiting to be executed
        running:
            Model run is running
        stopped:
            Model run was stopped (killed) by user
        done:
            Model run was completed succesfully
        failed:
            Model run completed running with an exit code
        """
        if self._status[model_run_name] == 'running':
            if self.lock is False:
                self.lock = True
                for line in iter(self._process[model_run_name].stdout.readline, b''):
                    self._output[model_run_name] += line.decode()
                    self._process[model_run_name].stdout.flush()
                self.lock = False

            if self._process[model_run_name].poll() == 0:
                self._status[model_run_name] = 'done'
            elif self._process[model_run_name].poll() == 1:
                self._status[model_run_name] = 'failed'

        return {
            'status': self._status[model_run_name],
            'output': self._output[model_run_name]
        }


class JobScheduler(object):
    """Run JobGraphs produced by a :class:`~smif.controller.modelrun.ModelRun`
    """
    def __init__(self):
        self._status = defaultdict(lambda: 'unstarted')
        self._id_counter = itertools.count()
        self.logger = logging.getLogger(__name__)
        self.store = None

    def add(self, job_graph):
        """Add a JobGraph to the JobScheduler and run directly

        Arguments
        ---------
        job_graph: :class:`networkx.graph`
        """
        job_graph_id = self._next_id()
        try:
            self._run(job_graph, job_graph_id)
        except Exception as ex:
            self._status[job_graph_id] = 'failed'
            traceback.print_exc()
            return job_graph_id, ex

        return job_graph_id, None

    def kill(self, job_graph_id):
        """Kill a job_graph that is already running - not implemented

        Parameters
        ----------
        job_graph_id: int
        """
        raise NotImplementedError

    def get_status(self, job_graph_id):
        """Get job graph status

        Parameters
        ----------
        job_graph_id: int

        Returns
        -------
        dict: A message containing the status

        Notes
        -----
        Possible statuses:

        unstarted:
            Job graph has not yet started
        running:
            Job graph is running
        done:
            Job graph was completed succesfully
        failed:
            Job graph completed running with an exit code
        """
        return {'status': self._status[job_graph_id]}

    def _run(self, job_graph, job_graph_id):
        """Run a job graph
        - sort the jobs into a single list
        - unpack model, data_handle and operation from each node
        """
        self.logger.profiling_start('JobScheduler._run()', 'graph_' + str(job_graph_id))
        self._status[job_graph_id] = 'running'

        for job_node_id, job in self._get_run_order(job_graph):
            self.logger.info("Job %s", job_node_id)
            self.logger.profiling_start('JobScheduler._run()', 'job_' + job_node_id)

            model = job['model']
            data_handle = DataHandle(
                store=self.store,
                model=model,
                modelrun_name=job['modelrun_name'],
                current_timestep=job['current_timestep'],
                timesteps=job['timesteps'],
                decision_iteration=job['decision_iteration']
            )
            operation = job['operation']
            if operation is ModelOperation.BEFORE_MODEL_RUN:
                # before_model_run may not be implemented by all jobs
                if hasattr(model, "before_model_run"):
                    model.before_model_run(data_handle)

            elif operation is ModelOperation.SIMULATE:
                model.simulate(data_handle)

            else:
                raise ValueError("Unrecognised operation: {}".format(operation))
            self.logger.profiling_stop('JobScheduler._run()', 'job_' + job_node_id)

        self._status[job_graph_id] = 'done'
        self.logger.profiling_stop('JobScheduler._run()', 'graph_' + str(job_graph_id))

    def _next_id(self):
        return next(self._id_counter)

    @staticmethod
    def _get_run_order(graph):
        """Returns a list of jobs in a runnable order.

        Returns
        -------
        list
            A list of job nodes
        """
        try:
            # topological sort gives a single list from directed graph,
            # ignoring opportunities to run independent models in parallel
            run_order = networkx.topological_sort(graph)

            # list of Models (typically ScenarioModel and SectorModel)
            ordered_jobs = [(run, graph.nodes[run]) for run in run_order]
        except networkx.NetworkXUnfeasible:
            raise NotImplementedError("Job graphs must not contain cycles")

        return ordered_jobs



from copy import copy
from unittest.mock import Mock

from pytest import fixture, raises
from smif.controller.modelrun import ModelRunBuilder, ModelRunner
from smif.exception import SmifModelRunError
from smif.metadata import RelativeTimestep, Spec
from smif.model import ScenarioModel, SectorModel, SosModel


class EmptySectorModel(SectorModel):
    def simulate(self, data):
        return data


@fixture(scope='function')
def config_data():
    """Config for a model run
    """
    sos_model = SosModel('sos_model')

    climate_scenario = ScenarioModel('climate')
    sos_model.add_model(climate_scenario)

    energy_supply = EmptySectorModel('energy_supply')
    sos_model.add_model(energy_supply)

    config = {
        'name': 'unique_model_run_name',
        'stamp': '2017-09-20T12:53:23+00:00',
        'description': 'a description of what the model run contains',
        'timesteps': [2010, 2011, 2012],
        'sos_model': sos_model,
        'scenarios': {
            'climate': 'RCP4.5'
        },
        'narratives': [
        ],
        'strategies': [
            {
                'strategy': 'pre-specified-planning',
                'description': 'build_nuclear',
                'model_name': 'energy_supply',
                'interventions': [
                    {'name': 'nuclear_large', 'build_year': 2012},
                    {'name': 'carrington_retire', 'build_year': 2011}
                ]
            }
        ]
    }
    return config


@fixture(scope='function')
def model_run(config_data):
    """ModelRun built from config
    """
    builder = ModelRunBuilder()
    builder.construct(config_data)
    return builder.finish()


@fixture(scope='function')
def mock_model_run():
    """Minimal mock ModelRun
    """
    sos_model = SosModel('test_sos_model')
    sos_model.parameters = {}

    modelrun = Mock()
    modelrun.name = 'test'
    modelrun.strategies = []
    modelrun.sos_model = sos_model
    modelrun.narratives = {}
    modelrun.model_horizon = [1]
    modelrun.initialised = False
    return modelrun


@fixture(scope='function')
def mock_store():
    """Minimal mock store
    """
    store = Mock()
    store.read_model_run = Mock(return_value={
        'sos_model': 'test_sos_model',
        'narratives': {},
        'scenarios': {}
    })
    store.read_sos_model = Mock(return_value={
        'name': 'test_sos_model',
        'model_dependencies': [],
        'scenario_dependencies': [],
        'sector_models': ['sector_model_test']
    })
    store.read_strategies = Mock(return_value=[])
    store.read_all_initial_conditions = Mock(return_value=[])
    store.read_interventions = Mock(return_value={})

    return store


class TestModelRunBuilder:
    """Build from config
    """
    def test_builder(self, config_data):
        """Test basic properties
        """
        builder = ModelRunBuilder()
        builder.construct(config_data)
        modelrun = builder.finish()

        assert modelrun.name == 'unique_model_run_name'
        assert modelrun.timestamp == '2017-09-20T12:53:23+00:00'
        assert modelrun.model_horizon == [2010, 2011, 2012]
        assert modelrun.status == 'Built'
        assert modelrun.scenarios == {'climate': 'RCP4.5'}
        assert modelrun.narratives == config_data['narratives']
        assert modelrun.strategies == config_data['strategies']

    def test_builder_scenario_sosmodelrun_not_in_sosmodel(self, config_data):
        """Error from unused scenarios
        """
        config_data['scenarios'] = {
            'climate': 'RCP4.5',
            'population': 'high_population'
        }
        builder = ModelRunBuilder()
        builder.construct(config_data)

        with raises(SmifModelRunError) as ex:
            builder.finish()
        assert "ScenarioSets {'population'} are selected in the ModelRun " \
               "configuration but not found in the SosModel configuration" in str(ex)


class TestModelRun:
    """Core ModelRun
    """
    def test_run_static(self, model_run, mock_store):
        """Call run
        """
        model_run.run(mock_store)

    def test_run_timesteps(self, config_data):
        """should error that timesteps are empty
        """
        config_data['timesteps'] = []
        builder = ModelRunBuilder()
        builder.construct(config_data)
        model_run = builder.finish()
        store = Mock()
        with raises(SmifModelRunError) as ex:
            model_run.run(store)
        assert 'No timesteps specified' in str(ex)

    def test_serialize(self, config_data):
        """Serialise back to config dict
        """
        builder = ModelRunBuilder()
        builder.construct(config_data)
        model_run = builder.finish()

        expected = copy(config_data)
        expected['sos_model'] = config_data['sos_model'].name  # expect a reference by name
        assert expected == model_run.as_dict()


class TestModelRunnerJobGraphs():
    """Cover all JobGraph corner cases
    """
    def test_jobgraph_single_timestep(self, mock_model_run):
        """
        a[before]
        |
        v
        a[sim]
        """
        model_a = EmptySectorModel('model_a')
        mock_model_run.sos_model.add_model(model_a)

        runner = ModelRunner()
        bundle = {
            'decision_iterations': [0],
            'timesteps': [1]
        }
        job_graph = runner.build_job_graph(mock_model_run, bundle)
        print(job_graph.nodes, job_graph.edges)

        actual = list(job_graph.predecessors('test_before_model_run_model_a'))
        expected = []
        assert actual == expected

        actual = list(job_graph.successors('test_before_model_run_model_a'))
        expected = ['test_simulate_1_0_model_a']
        assert actual == expected

        actual = list(job_graph.predecessors('test_simulate_1_0_model_a'))
        expected = ['test_before_model_run_model_a']
        assert actual == expected

        actual = list(job_graph.successors('test_simulate_1_0_model_a'))
        expected = []
        assert actual == expected

    def test_jobgraph_multiple_timesteps(self, mock_model_run):
        """
        a[before]
        |        |
        v        V
        a[sim]  a[sim]
        t=1     t=2
        """
        model_a = EmptySectorModel('model_a')
        mock_model_run.sos_model.add_model(model_a)

        mock_model_run.model_horizon = [1, 2]

        runner = ModelRunner()
        bundle = {
            'decision_iterations': [0],
            'timesteps': [1, 2]
        }
        job_graph = runner.build_job_graph(mock_model_run, bundle)

        actual = list(job_graph.predecessors('test_before_model_run_model_a'))
        expected = []
        assert actual == expected

        actual = list(job_graph.successors('test_before_model_run_model_a'))
        expected = ['test_simulate_1_0_model_a', 'test_simulate_2_0_model_a']
        assert sorted(actual) == sorted(expected)

        actual = list(job_graph.predecessors('test_simulate_1_0_model_a'))
        expected = ['test_before_model_run_model_a']
        assert actual == expected

        actual = list(job_graph.successors('test_simulate_1_0_model_a'))
        expected = []
        assert actual == expected

        actual = list(job_graph.predecessors('test_simulate_2_0_model_a'))
        expected = ['test_before_model_run_model_a']
        assert actual == expected

        actual = list(job_graph.successors('test_simulate_2_0_model_a'))
        expected = []
        assert actual == expected

    def test_jobgraph_multiple_timesteps_with_dep(self, mock_model_run):
        """
        a[before]
        |        |
        v        V
        a[sim]   a[sim]
        t=1 ---> t=2
        """
        model_a = EmptySectorModel('model_a')
        model_a.add_input(Spec('input', dtype='float'))
        model_a.add_output(Spec('output', dtype='float'))

        mock_model_run.sos_model.add_model(model_a)
        mock_model_run.sos_model.add_dependency(
            model_a, 'output',
            model_a, 'input',
            RelativeTimestep.PREVIOUS)

        mock_model_run.model_horizon = [1, 2]

        runner = ModelRunner()
        bundle = {
            'decision_iterations': [0],
            'timesteps': [1, 2]
        }
        job_graph = runner.build_job_graph(mock_model_run, bundle)

        actual = list(job_graph.predecessors('test_simulate_1_0_model_a'))
        expected = ['test_before_model_run_model_a']
        assert actual == expected

        actual = list(job_graph.predecessors('test_simulate_2_0_model_a'))
        expected = ['test_before_model_run_model_a', 'test_simulate_1_0_model_a']
        assert sorted(actual) == sorted(expected)

    def test_jobgraph_multiple_models(self, mock_model_run):
        """
        a[before]   b[before]   c[before]
        |           |           |
        v           V           V
        a[sim] ---> b[sim] ---> c[sim]
           |------------------>
        """
        model_a = EmptySectorModel('model_a')
        model_a.add_output(Spec('a', dtype='float'))

        model_b = EmptySectorModel('model_b')
        model_b.add_input(Spec('a', dtype='float'))
        model_b.add_output(Spec('b', dtype='float'))

        model_c = EmptySectorModel('model_c')
        model_c.add_input(Spec('a', dtype='float'))
        model_c.add_input(Spec('b', dtype='float'))

        mock_model_run.sos_model.add_model(model_a)
        mock_model_run.sos_model.add_model(model_b)
        mock_model_run.sos_model.add_model(model_c)

        mock_model_run.sos_model.add_dependency(model_a, 'a', model_b, 'a')
        mock_model_run.sos_model.add_dependency(model_a, 'a', model_c, 'a')
        mock_model_run.sos_model.add_dependency(model_b, 'b', model_c, 'b')

        runner = ModelRunner()
        bundle = {
            'decision_iterations': [0],
            'timesteps': [1]
        }
        job_graph = runner.build_job_graph(mock_model_run, bundle)

        actual = list(job_graph.predecessors('test_simulate_1_0_model_a'))
        expected = ['test_before_model_run_model_a']
        assert actual == expected

        actual = list(job_graph.successors('test_simulate_1_0_model_a'))
        expected = ['test_simulate_1_0_model_b', 'test_simulate_1_0_model_c']
        assert sorted(actual) == sorted(expected)

        actual = list(job_graph.predecessors('test_simulate_1_0_model_b'))
        expected = ['test_before_model_run_model_b', 'test_simulate_1_0_model_a']
        assert sorted(actual) == sorted(expected)

        actual = list(job_graph.successors('test_simulate_1_0_model_b'))
        expected = ['test_simulate_1_0_model_c']
        assert actual == expected

        actual = list(job_graph.predecessors('test_simulate_1_0_model_c'))
        expected = ['test_before_model_run_model_c', 'test_simulate_1_0_model_a',
                    'test_simulate_1_0_model_b']
        assert sorted(actual) == sorted(expected)

        actual = list(job_graph.successors('test_simulate_1_0_model_c'))
        expected = []
        assert actual == expected

    def test_jobgraph_interdependency(self, mock_model_run):
        """
        a[before]   b[before]
        |           |
        v           V
        a[sim] ---> b[sim]
               <---
        """
        model_a = EmptySectorModel('model_a')
        model_a.add_input(Spec('b', dtype='float'))
        model_a.add_output(Spec('a', dtype='float'))

        model_b = EmptySectorModel('model_b')
        model_b.add_input(Spec('a', dtype='float'))
        model_b.add_output(Spec('b', dtype='float'))

        mock_model_run.sos_model.add_model(model_a)
        mock_model_run.sos_model.add_model(model_b)

        mock_model_run.sos_model.add_dependency(model_a, 'a', model_b, 'a')
        mock_model_run.sos_model.add_dependency(model_b, 'b', model_a, 'b')

        runner = ModelRunner()
        bundle = {
            'decision_iterations': [0],
            'timesteps': [1]
        }
        with raises(NotImplementedError):
            runner.build_job_graph(mock_model_run, bundle)

    def test_jobgraph_with_models_initialised(self, mock_model_run):
        """
        a[sim]
        """
        model_a = EmptySectorModel('model_a')
        mock_model_run.sos_model.add_model(model_a)
        mock_model_run.initialised = True

        runner = ModelRunner()
        bundle = {
            'decision_iterations': [0],
            'timesteps': [1]
        }
        job_graph = runner.build_job_graph(mock_model_run, bundle)

        actual = list(job_graph.predecessors('test_simulate_1_0_model_a'))
        expected = []
        assert actual == expected

        actual = list(job_graph.successors('test_simulate_1_0_model_a'))
        expected = []
        assert actual == expected



"""Test YAML config store
"""
from pytest import fixture, raises
from smif.data_layer.file.file_config_store import YamlConfigStore
from smif.exception import (SmifDataExistsError, SmifDataMismatchError,
                            SmifDataNotFoundError)


@fixture(scope='function')
def config_handler(setup_folder_structure, get_sector_model, sample_scenarios):
    handler = YamlConfigStore(str(setup_folder_structure), validation=False)
    for scenario in sample_scenarios:
        handler.write_scenario(scenario)
    handler.write_model(get_sector_model)
    return handler


class TestModelRun:
    """Model runs should be defined once, hard to overwrite
    """
    def test_model_run_read_all(self, model_run, config_handler):
        """Test to write two model_run configurations to Yaml files, then
        read the Yaml files and compare that the result is equal.
        """

        model_run1 = model_run
        model_run1['name'] = 'model_run1'
        config_handler.write_model_run(model_run1)

        model_run2 = model_run
        model_run2['name'] = 'model_run2'
        config_handler.write_model_run(model_run2)

        model_runs = config_handler.read_model_runs()
        model_run_names = list(model_run['name'] for model_run in model_runs)

        assert 'model_run1' in model_run_names
        assert 'model_run2' in model_run_names
        assert len(model_runs) == 2

    def test_model_run_write_twice(self, model_run, config_handler):
        """Test that writing a model_run should fail (not overwrite).
        """

        model_run1 = model_run
        model_run1['name'] = 'unique'
        config_handler.write_model_run(model_run1)

        with raises(SmifDataExistsError) as ex:
            config_handler.write_model_run(model_run1)
        assert "Model_run 'unique' already exists" in str(ex)

    def test_model_run_read_one(self, model_run, config_handler):
        """Test reading a single model_run.
        """

        model_run1 = model_run
        model_run1['name'] = 'model_run1'
        config_handler.write_model_run(model_run1)

        model_run2 = model_run
        model_run2['name'] = 'model_run2'
        config_handler.write_model_run(model_run2)

        model_run = config_handler.read_model_run('model_run2')
        assert model_run['name'] == 'model_run2'

    def test_model_run_read_missing(self, config_handler):
        """Test that reading a missing model_run fails.
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.read_model_run('missing_name')
        assert "Model_run 'missing_name' not found" in str(ex)

    def test_model_run_update(self, model_run, config_handler):
        """Test updating a model_run description
        """
        model_run = model_run
        model_run['name'] = 'to_update'
        model_run['description'] = 'before'

        config_handler.write_model_run(model_run)

        model_run['description'] = 'after'
        config_handler.update_model_run('to_update', model_run)

        actual = config_handler.read_model_run('to_update')
        assert actual['description'] == 'after'

    def test_model_run_update_mismatch(self, model_run, config_handler):
        """Test that updating a model_run with mismatched name should fail
        """
        model_run = model_run

        model_run['name'] = 'model_run'
        with raises(SmifDataMismatchError) as ex:
            config_handler.update_model_run('model_run2', model_run)
        assert "name 'model_run2' must match 'model_run'" in str(ex)

    def test_model_run_update_missing(self, model_run, config_handler):
        """Test that updating a nonexistent model_run should fail
        """
        model_run = model_run
        model_run['name'] = 'missing_name'

        with raises(SmifDataNotFoundError) as ex:
            config_handler.update_model_run('missing_name', model_run)
        assert "Model_run 'missing_name' not found" in str(ex)

    def test_model_run_delete(self, model_run, config_handler):
        """Test that updating a nonexistent model_run should fail
        """
        model_run = model_run
        model_run['name'] = 'to_delete'

        config_handler.write_model_run(model_run)
        before_delete = config_handler.read_model_runs()
        assert len(before_delete) == 1

        config_handler.delete_model_run('to_delete')
        after_delete = config_handler.read_model_runs()
        assert len(after_delete) == 0

    def test_model_run_delete_missing(self, model_run, config_handler):
        """Test that updating a nonexistent model_run should fail
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.delete_model_run('missing_name')
        assert "Model_run 'missing_name' not found" in str(ex)


class TestSosModel:
    """SoSModel configurations should be accessible and editable.
    """
    def test_sos_model_read_all(self, get_sos_model, config_handler):
        """Test to write two sos_model configurations to Yaml files, then
        read the Yaml files and compare that the result is equal.
        """

        sos_model1 = get_sos_model
        sos_model1['name'] = 'sos_model1'
        config_handler.write_sos_model(sos_model1)

        sos_model2 = get_sos_model
        sos_model2['name'] = 'sos_model2'
        config_handler.write_sos_model(sos_model2)

        sos_models = config_handler.read_sos_models()
        sos_model_names = list(sos_model['name'] for sos_model in sos_models)

        assert 'sos_model1' in sos_model_names
        assert 'sos_model2' in sos_model_names
        assert len(sos_models) == 2

    def test_sos_model_write_twice(self, get_sos_model, config_handler):
        """Test that writing a sos_model should fail (not overwrite).
        """

        sos_model1 = get_sos_model
        sos_model1['name'] = 'unique'
        config_handler.write_sos_model(sos_model1)

        with raises(SmifDataExistsError) as ex:
            config_handler.write_sos_model(sos_model1)
        assert "Sos_model 'unique' already exists" in str(ex)

    def test_sos_model_read_one(self, get_sos_model, config_handler):
        """Test reading a single sos_model.
        """

        sos_model1 = get_sos_model
        sos_model1['name'] = 'sos_model1'
        config_handler.write_sos_model(sos_model1)

        sos_model2 = get_sos_model
        sos_model2['name'] = 'sos_model2'
        config_handler.write_sos_model(sos_model2)

        sos_model = config_handler.read_sos_model('sos_model2')
        assert sos_model['name'] == 'sos_model2'

    def test_sos_model_read_missing(self, config_handler):
        """Test that reading a missing sos_model fails.
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.read_sos_model('missing_name')
        assert "Sos_model 'missing_name' not found" in str(ex)

    def test_sos_model_update(self, get_sos_model, config_handler):
        """Test updating a sos_model description
        """
        sos_model = get_sos_model
        sos_model['name'] = 'to_update'
        sos_model['description'] = 'before'

        config_handler.write_sos_model(sos_model)

        sos_model['description'] = 'after'
        config_handler.update_sos_model('to_update', sos_model)

        actual = config_handler.read_sos_model('to_update')
        assert actual['description'] == 'after'

    def test_sos_model_update_mismatch(self, get_sos_model, config_handler):
        """Test that updating a sos_model with mismatched name should fail
        """
        sos_model = get_sos_model

        sos_model['name'] = 'sos_model'
        with raises(SmifDataMismatchError) as ex:
            config_handler.update_sos_model('sos_model2', sos_model)
        assert "name 'sos_model2' must match 'sos_model'" in str(ex)

    def test_sos_model_update_missing(self, get_sos_model, config_handler):
        """Test that updating a nonexistent sos_model should fail
        """
        sos_model = get_sos_model
        sos_model['name'] = 'missing_name'

        with raises(SmifDataNotFoundError) as ex:
            config_handler.update_sos_model('missing_name', sos_model)
        assert "Sos_model 'missing_name' not found" in str(ex)

    def test_sos_model_delete(self, get_sos_model, config_handler):
        """Test that updating a nonexistent sos_model should fail
        """
        sos_model = get_sos_model
        sos_model['name'] = 'to_delete'

        config_handler.write_sos_model(sos_model)
        before_delete = config_handler.read_sos_models()
        assert len(before_delete) == 1

        config_handler.delete_sos_model('to_delete')
        after_delete = config_handler.read_sos_models()
        assert len(after_delete) == 0

    def test_sos_model_delete_missing(self, get_sos_model, config_handler):
        """Test that updating a nonexistent sos_model should fail
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.delete_sos_model('missing_name')
        assert "Sos_model 'missing_name' not found" in str(ex)


class TestSectorModel:
    """SectorModel definitions should be accessible - may move towards being less editable
    as config, more defined in code/wrapper.
    """
    def test_sector_model_read_all(self, get_sector_model, config_handler):
        """Test to write two sector_model configurations to Yaml files, then
        read the Yaml files and compare that the result is equal.
        """

        sector_model1 = get_sector_model
        sector_model1['name'] = 'sector_model1'
        config_handler.write_model(sector_model1)

        sector_model2 = get_sector_model
        sector_model2['name'] = 'sector_model2'
        config_handler.write_model(sector_model2)

        sector_models = config_handler.read_models()
        sector_model_names = list(sector_model['name'] for sector_model in sector_models)

        assert 'sector_model1' in sector_model_names
        assert 'sector_model2' in sector_model_names
        assert 'energy_demand' in sector_model_names
        assert len(sector_models) == 3

    def test_sector_model_write_twice(self, get_sector_model, config_handler):
        """Test that writing a sector_model should fail (not overwrite).
        """

        sector_model1 = get_sector_model
        sector_model1['name'] = 'unique'
        config_handler.write_model(sector_model1)

        with raises(SmifDataExistsError) as ex:
            config_handler.write_model(sector_model1)
        assert "Sector_model 'unique' already exists" in str(ex)

    def test_sector_model_read_one(self, get_sector_model, config_handler):
        """Test reading a single sector_model.
        """

        sector_model1 = get_sector_model
        sector_model1['name'] = 'sector_model1'
        config_handler.write_model(sector_model1)

        sector_model2 = get_sector_model
        sector_model2['name'] = 'sector_model2'
        config_handler.write_model(sector_model2)

        sector_model = config_handler.read_model('sector_model2')
        assert sector_model['name'] == 'sector_model2'

    def test_sector_model_read_missing(self, config_handler):
        """Test that reading a missing sector_model fails.
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.read_model('missing_name')
        assert "Sector_model 'missing_name' not found" in str(ex)

    def test_sector_model_update(self, get_sector_model, config_handler):
        """Test updating a sector_model description
        """
        sector_model = get_sector_model
        sector_model['name'] = 'to_update'
        sector_model['description'] = 'before'

        config_handler.write_model(sector_model)

        sector_model['description'] = 'after'
        config_handler.update_model('to_update', sector_model)

        actual = config_handler.read_model('to_update')
        assert actual['description'] == 'after'

    def test_sector_model_update_mismatch(self, get_sector_model, config_handler):
        """Test that updating a sector_model with mismatched name should fail
        """
        sector_model = get_sector_model
        sector_model['name'] = 'sector_model2'
        config_handler.write_model(sector_model)

        sector_model['name'] = 'sector_model'
        with raises(SmifDataMismatchError) as ex:
            config_handler.update_model('sector_model2', sector_model)
        assert "name 'sector_model2' must match 'sector_model'" in str(ex)

    def test_sector_model_update_missing(self, get_sector_model, config_handler):
        """Test that updating a nonexistent sector_model should fail
        """
        sector_model = get_sector_model
        sector_model['name'] = 'missing_name'

        with raises(SmifDataNotFoundError) as ex:
            config_handler.update_model('missing_name', sector_model)
        assert "Sector_model 'missing_name' not found" in str(ex)

    def test_sector_model_delete(self, get_sector_model, config_handler):
        """Test that updating a nonexistent sector_model should fail
        """
        before_delete = config_handler.read_models()
        assert len(before_delete) == 1

        config_handler.delete_model('energy_demand')
        after_delete = config_handler.read_models()
        assert len(after_delete) == 0

    def test_sector_model_delete_missing(self, get_sector_model, config_handler):
        """Test that updating a nonexistent sector_model should fail
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.delete_model('missing_name')
        assert "Sector_model 'missing_name' not found" in str(ex)


class TestScenarios:
    """Scenario data should be readable, metadata is currently editable. May move to make it
    possible to import/edit/write data.
    """
    def test_read_scenario_definition(self, setup_folder_structure, config_handler,
                                      sample_scenarios):
        """Should read a scenario definition
        """
        expected = sample_scenarios[0]
        actual = config_handler.read_scenario(expected['name'])
        assert actual == expected

    def test_read_scenario_missing(self, config_handler):
        """Should raise a SmifDataNotFoundError if scenario not found
        """
        with raises(SmifDataNotFoundError) as ex:
            config_handler.read_scenario('missing')
        assert "Scenario 'missing' not found" in str(ex)



"""Test all DataStore implementations
"""
import numpy as np
from pytest import fixture, mark, param, raises
from smif.data_layer.data_array import DataArray
from smif.data_layer.database_interface import DbDataStore
from smif.data_layer.file.file_data_store import CSVDataStore, ParquetDataStore
from smif.data_layer.memory_interface import MemoryDataStore
from smif.exception import SmifDataNotFoundError
from smif.metadata import Spec


@fixture(
    params=[
        'memory',
        'file_csv',
        'file_parquet',
        param('database', marks=mark.skip)]
    )
def handler(request, setup_empty_folder_structure):
    if request.param == 'memory':
        handler = MemoryDataStore()
    elif request.param == 'file_csv':
        base_folder = setup_empty_folder_structure
        handler = CSVDataStore(base_folder)
    elif request.param == 'file_parquet':
        base_folder = setup_empty_folder_structure
        handler = ParquetDataStore(base_folder)
    elif request.param == 'database':
        handler = DbDataStore()
        raise NotImplementedError

    return handler


class TestDataArray():
    """Read and write DataArray
    """
    def test_read_write_data_array(self, handler, scenario):
        data = np.array([0, 1], dtype=float)
        spec = Spec.from_dict(scenario['provides'][0])

        da = DataArray(spec, data)

        handler.write_scenario_variant_data('mortality.csv', da, 2010)
        actual = handler.read_scenario_variant_data('mortality.csv', spec, 2010)

        assert actual == da

    def test_read_zero_d_from_timeseries(self, handler):
        """Read a single value
            timestep,param
            2010,0
            2015,1
        """
        # write data for multiple timesteps by adding 'timestep' dim to the spec
        data = np.array([0, 1], dtype=float)
        write_spec = Spec(
            name='param',
            dims=['timestep'],
            coords={'timestep': [2010, 2015]},
            dtype='float'
        )
        da = DataArray(write_spec, data)
        handler.write_scenario_variant_data('param', da)

        read_spec = Spec(
            name='param',
            dims=[],
            coords={},
            dtype='float'
        )
        actual = handler.read_scenario_variant_data('param', read_spec, 2010).data
        assert actual == np.array(0.0)
        actual = handler.read_scenario_variant_data('param', read_spec, 2015).data
        assert actual == np.array(1.0)

    def test_read_data_array_missing_timestep(self, handler, scenario):
        data = np.array([0, 1], dtype=float)
        spec = Spec.from_dict(scenario['provides'][0])

        da = DataArray(spec, data)

        handler.write_scenario_variant_data('mortality.csv', da, 2010)
        msg = "not found for timestep 2011"
        with raises(SmifDataNotFoundError) as ex:
            handler.read_scenario_variant_data('mortality.csv', spec, 2011)
        assert msg in str(ex)

    def test_string_data(self, handler):
        spec = Spec(
            name='string_data',
            dims=['zones'],
            coords={'zones': ['a', 'b', 'c']},
            dtype='object'
        )
        data = np.array(['alpha', 'beta', 'γάμμα'], dtype='object')
        expected = DataArray(spec, data)

        handler.write_scenario_variant_data('key', expected, 2010)
        actual = handler.read_scenario_variant_data('key', spec, 2010)
        assert actual == expected


class TestInitialConditions():
    """Read and write initial conditions
    """
    def test_read_write_initial_conditions(self, handler, initial_conditions):

        expected = initial_conditions

        handler.write_initial_conditions('initial_conditions.csv', initial_conditions)
        actual = handler.read_initial_conditions(['initial_conditions.csv'])

        assert actual == expected


class TestInterventions():
    """Read and write interventions
    """
    def test_read_write_interventions(self, handler, interventions):

        expected = interventions
        handler.write_interventions('my_intervention.csv', interventions)
        actual = handler.read_interventions(['my_intervention.csv'])

        assert actual == expected


class TestState():
    """Read and write state
    """
    def test_read_write_state(self, handler, state):
        expected = state
        modelrun_name = 'test_modelrun'
        timestep = 2020
        decision_iteration = None

        handler.write_state(expected, modelrun_name, timestep, decision_iteration)
        actual = handler.read_state(modelrun_name, timestep, decision_iteration)
        assert actual == expected


class TestCoefficients():
    """Read/write conversion coefficients
    """
    def test_read_write_coefficients(self, handler):

        with raises(SmifDataNotFoundError):
            handler.read_coefficients('from_dim_name', 'to_dim_name')

        expected = np.array([[2]])
        handler.write_coefficients('from_dim_name', 'to_dim_name', expected)

        actual = handler.read_coefficients('from_dim_name', 'to_dim_name')
        np.testing.assert_equal(actual, expected)


class TestResults():
    """Read/write results and prepare warm start
    """
    def test_read_write_results(self, handler, sample_results):
        output_spec = sample_results.spec
        modelrun_name = 'test_modelrun'
        model_name = 'energy'
        timestep = 2010

        handler.write_results(sample_results, modelrun_name, model_name, timestep)
        results_out = handler.read_results(modelrun_name, model_name, output_spec, timestep)

        assert results_out == sample_results

    def test_available_results(self, handler, sample_results):
        """Available results should return an empty list if none are available
        develop
        """
        assert handler.available_results('test_modelrun') == []
        handler.write_results(sample_results, 'test_modelrun', 'energy', 2010, 0)

        # keys should be (timestep, decision_iteration, model_name, output_name)
        assert handler.available_results('test_modelrun') == \
            [(2010, 0, 'energy', sample_results.spec.name)]

    def test_read_results_raises(self, handler, sample_results):
        modelrun_name = 'test_modelrun'
        model_name = 'energy'
        timestep = 2010
        output_spec = sample_results.spec

        handler.write_results(sample_results, modelrun_name, model_name, timestep=timestep)

        with raises(SmifDataNotFoundError):
            handler.read_results(modelrun_name, model_name, output_spec, 2020)



# -*- coding: utf-8 -*-
"""This module coordinates the software components that make up the integration
framework.

A system of systems model contains simulation and scenario models,
and the dependencies between the models.

"""
import itertools
import logging
from collections import defaultdict

from smif.exception import SmifDataMismatchError, SmifValidationError
from smif.metadata import RelativeTimestep
from smif.model.dependency import Dependency
from smif.model.model import Model, ScenarioModel
from smif.model.sector_model import SectorModel

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class SosModel():
    """Consists of the collection of models joined via dependencies

    Arguments
    ---------
    name : str
        The unique name of the SosModel

    """
    def __init__(self, name):
        self.logger = logging.getLogger(__name__)

        self.name = name
        self.description = ''

        # Models in an internal lookup by name
        # { name: Model }
        self._models = {}

        # Maintain dependencies in an internal lookup (using names (str) in the tuple key)
        # { (source, output, sink, input): list[Dependency]}
        self._scenario_dependencies = defaultdict(list)
        self._model_dependencies = defaultdict(list)

        self.narratives = {}

    def as_dict(self):
        """Serialize the SosModel object

        Returns
        -------
        dict
        """
        scenario_dependencies = []
        for dep in self._scenario_dependencies.values():
            scenario_dependencies.append(dep.as_dict())

        model_dependencies = []
        for dep in self._model_dependencies.values():
            model_dependencies.append(dep.as_dict())

        config = {
            'name': self.name,
            'description': self.description,
            'scenarios': sorted(
                scenario.name
                for scenario in self.scenario_models
            ),
            'sector_models': sorted(
                model.name
                for model in self.sector_models
            ),
            'scenario_dependencies': scenario_dependencies,
            'model_dependencies': model_dependencies,
            'narratives': self.narratives
        }

        return config

    @classmethod
    def from_dict(cls, data, models=None):
        """Create a SosModel from config data

        Arguments
        ---------
        data: dict
            Configuration data. Must include name. May include description. If models are
            provided, must include dependencies.
        models: list of Model
            Optional. If provided, must include each ScenarioModel and SectorModel referred to
            in the data['dependencies']
        """
        sos_model = cls(data['name'])

        try:
            sos_model.description = data['description']
        except KeyError:
            pass

        if models:
            for model in models:
                sos_model.add_model(model)

            for dep in data['model_dependencies'] + data['scenario_dependencies']:
                sink = SosModel._get_dependency(sos_model, dep, 'sink')
                source = SosModel._get_dependency(sos_model, dep, 'source')
                source_output_name = dep['source_output']
                sink_input_name = dep['sink_input']
                try:
                    timestep = RelativeTimestep.from_name(dep['timestep'])
                except KeyError:
                    # if not provided, default to current
                    timestep = RelativeTimestep.CURRENT
                except ValueError:
                    # if not parseable as relative timestep, pass through
                    pass

                sos_model.add_dependency(
                    source, source_output_name, sink, sink_input_name, timestep)

        sos_model.check_dependencies()
        return sos_model

    @staticmethod
    def _get_dependency(sos_model, dep, source_or_sink):
        try:
            source = sos_model.get_model(dep[source_or_sink])
        except KeyError:
            msg = 'SectorModel or ScenarioModel {} `{}` required by ' + \
                  'dependency `{}` was not provided by the builder'
            dependency = (
                dep['source'] + ' (' + dep['source_output'] + ')' +
                ' - ' +
                dep['sink'] + ' (' + dep['sink_input'] + ')'
            )
            raise SmifDataMismatchError(msg.format(source_or_sink,
                                                   dep[source_or_sink],
                                                   dependency))
        return source

    @property
    def models(self):
        """The models contained within this system-of-systems model

        Returns
        -------
        list[smif.model.model.Model]
        """
        return list(self._models.values())

    @property
    def sector_models(self):
        """Sector model objects contained in the SosModel

        Returns
        =======
        list
            A list of sector model objects
        """
        return [
            model for model in self.models
            if isinstance(model, SectorModel)
        ]

    @property
    def scenario_models(self):
        """Scenario model objects contained in the SosModel

        Returns
        -------
        list
            A list of scenario model objects
        """
        return [
            model for model in self.models
            if isinstance(model, ScenarioModel)
        ]

    def add_model(self, model):
        """Adds a model to the system-of-systems model

        Arguments
        ---------
        model : :class:`smif.model.model.Model`
        """
        msg = "Only Models can be added to a SosModel (and SosModels cannot be nested)"
        assert isinstance(model, Model), msg
        self.logger.info("Loading model: %s", model.name)
        self._models[model.name] = model

    def get_model(self, model_name):
        """Get a model by name

        Arguments
        ---------
        model_name : str

        Returns
        -------
        smif.model.model.Model
        """
        return self._models[model_name]

    @property
    def dependencies(self):
        """Dependency connections between models within this system-of-systems model

        Returns
        -------
        iterable[smif.model.dependency.Dependency]
        """
        return itertools.chain(
            self._scenario_dependencies.values(),
            self._model_dependencies.values()
        )

    @property
    def scenario_dependencies(self):
        """Dependency connections between models within this system-of-systems model

        Returns
        -------
        iterable[smif.model.dependency.Dependency]
        """
        return self._scenario_dependencies.values()

    @property
    def model_dependencies(self):
        """Dependency connections between models within this system-of-systems model

        Returns
        -------
        iterable[smif.model.dependency.Dependency]
        """
        return self._model_dependencies.values()

    def add_dependency(self, source_model, source_output_name, sink_model, sink_input_name,
                       timestep=RelativeTimestep.CURRENT):
        """Adds a dependency to this system-of-systems model

        Arguments
        ---------
        source_model : smif.model.model.Model
            A reference to the source `~smif.model.model.Model` object
        source_output_name : string
            The name of the model_output defined in the `source_model`
        source_model : smif.model.model.Model
            A reference to the source `~smif.model.model.Model` object
        sink_name : string
            The name of a model_input defined in this object
        timestep : smif.metadata.RelativeTimestep, optional
            The relative timestep of the dependency, defaults to CURRENT, may be PREVIOUS.
        """
        try:
            self.get_model(source_model.name)
        except KeyError:
            msg = "Source model '{}' does not exist in list of models"
            raise SmifValidationError(msg.format(source_model.name))

        try:
            self.get_model(sink_model.name)
        except KeyError:
            msg = "Sink model '{}' does not exist in list of models"
            raise SmifValidationError(msg.format(sink_model.name))

        if source_output_name not in source_model.outputs:
            msg = "Output '{}' is not defined in '{}' model"
            raise SmifValidationError(msg.format(source_output_name, source_model.name))

        if sink_input_name not in sink_model.inputs:
            msg = "Input '{}' is not defined in '{}' model"
            raise SmifValidationError(msg.format(sink_input_name, sink_model.name))

        key = (source_model.name, source_output_name, sink_model.name, sink_input_name)

        if not self._allow_adding_dependency(
                source_model, source_output_name, sink_model, sink_input_name, timestep):
            msg = "Inputs: '%s'. Free inputs: '%s'."
            self.logger.debug(msg, sink_model.inputs, self.free_inputs)
            msg = "Could not add dependency: input '{}' already provided"
            raise SmifValidationError(msg.format(sink_input_name))

        source_spec = source_model.outputs[source_output_name]
        sink_spec = sink_model.inputs[sink_input_name]
        if isinstance(source_model, ScenarioModel):
            self._scenario_dependencies[key] = Dependency(
                source_model, source_spec, sink_model, sink_spec, timestep=timestep)
        else:
            self._model_dependencies[key] = Dependency(
                source_model, source_spec, sink_model, sink_spec, timestep=timestep)

        msg = "Added dependency from '%s:%s' to '%s:%s'"
        self.logger.debug(
            msg, source_model.name, source_output_name, self.name, sink_input_name)

    def _allow_adding_dependency(self, source_model, source_output_name, sink_model,
                                 sink_input_name, timestep):
        key = (source_model.name, source_output_name, sink_model.name, sink_input_name)
        existing_deps = key in self._scenario_dependencies or key in self._model_dependencies

        if existing_deps:
            if key in self._scenario_dependencies and key in self._model_dependencies:
                # allowed at most two sources
                allowable = False
            else:
                # if and only if one is a scenario and the other is to a previous timestep
                try:
                    other_dep = self._scenario_dependencies[key]
                except KeyError:
                    other_dep = self._model_dependencies[key]

                other_model = other_dep.source_model
                other_timestep = other_dep.timestep

                other_is_previous = other_timestep == RelativeTimestep.PREVIOUS
                other_is_scenario = isinstance(other_model, ScenarioModel)

                new_is_previous = timestep == RelativeTimestep.PREVIOUS
                new_is_scenario = isinstance(source_model, ScenarioModel)

                allowable = (new_is_previous and other_is_scenario) or \
                    (other_is_previous and new_is_scenario)
        else:
            allowable = True

        return allowable

    def check_dependencies(self):
        """For each contained model, compare dependency list against
        list of available models
        """
        if self.free_inputs:
            msg = "A SosModel must have all inputs linked to dependencies. " \
                  "Define dependencies for {}"
            raise NotImplementedError(msg.format(", ".join(
                str(key) for key in self.free_inputs.keys()
            )))

    @property
    def free_inputs(self):
        """Returns the free inputs to models which are not yet linked to a dependency.

        Returns
        -------
        dict of {(model_name, input_name): smif.metadata.Spec}
        """
        satisfied_inputs = set(
            (sink_name, sink_input_name)
            for (source_name, source_output_name, sink_name, sink_input_name), deps
            in itertools.chain(
                self._scenario_dependencies.items(), self._model_dependencies.items())
            if deps
        )
        all_inputs = set()
        all_specs = {}

        for model in self.models:
            try:
                for input_name, input_spec in model.inputs.items():
                    input_key = (model.name, input_name)
                    all_inputs.add(input_key)
                    all_specs[input_key] = input_spec
            except AttributeError:
                pass  # ScenarioModels don't have inputs

        unsatisfied_inputs = all_inputs - satisfied_inputs

        return {
            input_key: all_specs[input_key]
            for input_key in unsatisfied_inputs
        }

    @property
    def outputs(self):
        """All model outputs provided by models contained within this SosModel

        Returns
        -------
        dict of {(model_name, output_name): smif.metadata.Spec}
        """
        outputs = {}
        for model in self.models:
            for output_name, output_spec in model.outputs.items():
                outputs[(model.name, output_name)] = output_spec

        return outputs

    def add_narrative(self, narrative):
        """Add a narrative to the system-of-systems model

        Arguments
        ---------
        narrative: dict
        """
        try:
            name = narrative['name']
        except KeyError:
            msg = "Could not find narrative name. Narrative argument " \
                  "should be a dict. Received a '{}'."
            raise SmifDataMismatchError(msg.format(type(narrative)))

        for model, parameters in narrative['provides'].items():
            self._check_model_exists(model)
            for parameter in parameters:
                self._check_parameter_exists(parameter, model)

        self.narratives[name] = narrative

    def _check_model_exists(self, model_name):
        if model_name not in [model.name for model in self.models]:
            msg = "'{}' does not exist in '{}'"
            raise SmifDataMismatchError(msg.format(model_name, self.name))

    def _check_parameter_exists(self, parameter_name, model_name):
        model = self._models[model_name]
        if parameter_name not in model.parameters:
            msg = "Parameter '{}' does not exist in '{}'"
            raise SmifDataMismatchError(msg.format(parameter_name, model_name))



"""The Model Run collects scenarios, timesteps, narratives, and
model collection into a package which can be built and passed to
the ModelRunner to run.

The ModelRunner is responsible for running a ModelRun, including passing
in the correct data to the model between timesteps and calling to the
DecisionManager to obtain decisions.

ModeRun has attributes:
- id
- description
- sosmodel
- timesteps
- scenarios
- narratives
- strategy
- status

"""
from logging import getLogger

import networkx as nx
from smif.controller.scheduler import JobScheduler
from smif.decision.decision import DecisionManager
from smif.exception import SmifModelRunError, SmifTimestepResolutionError
from smif.metadata import RelativeTimestep
from smif.model import ModelOperation, ScenarioModel


class ModelRun(object):
    """Collects timesteps, scenarios , narratives and a SosModel together

    Attributes
    ----------
    name: str
        The unique name of the model run
    timestamp: :class:`datetime.datetime`
        An ISO8601 compatible timestamp of model run creation time
    description: str
        A friendly description of the model run
    sos_model: :class:`smif.model.sos_model.SosModel`
        The contained SosModel
    scenarios: dict
        For each scenario set, a mapping to a valid scenario within that set
    narratives: list
        A list of :class:`smif.parameters.Narrative` objects
    strategies: dict
    status: str
    logger: logging.Logger
    results: dict
    """

    def __init__(self):
        self.name = ""
        self.timestamp = None
        self.description = ""
        self.sos_model = None
        self._model_horizon = []

        self.scenarios = {}
        self.narratives = []
        self.strategies = None
        self.status = 'Empty'

        self.logger = getLogger(__name__)

        self.results = {}

    def as_dict(self):
        """Serialises :class:`smif.controller.modelrun.ModelRun`

        Returns a dictionary definition of a ModelRun which is
        equivalent to that required by :class:`smif.controller.modelrun.ModelRunBuilder`
        to construct a new model run

        Returns
        -------
        dict
        """
        config = {
            'name': self.name,
            'description': self.description,
            'stamp': self.timestamp,
            'timesteps': self._model_horizon,
            'sos_model': self.sos_model.name,
            'scenarios': self.scenarios,
            'narratives': self.narratives,
            'strategies': self.strategies
        }
        return config

    def validate(self):
        """Validate that this ModelRun has been set up with sufficient data
        to run
        """
        scenarios = set(self.scenarios)
        model_scenarios = set(scenario.name for scenario in self.sos_model.scenario_models)
        missing_scenarios = scenarios - model_scenarios
        if missing_scenarios:
            raise SmifModelRunError("ScenarioSets {} are selected in the ModelRun "
                                    "configuration but not found in the SosModel "
                                    "configuration".format(missing_scenarios))

    @property
    def model_horizon(self):
        """Returns the list of timesteps

        Returns
        =======
        list
            A list of timesteps, distinct and sorted in ascending order
        """
        return self._model_horizon.copy()

    @model_horizon.setter
    def model_horizon(self, value):
        self._model_horizon = sorted(list(set(value)))

    def run(self, store, warm_start_timestep=None):
        """Builds all the objects and passes them to the ModelRunner

        The idea is that this will add ModelRuns to a queue for asychronous
        processing

        """
        self.logger.debug("Running model run %s", self.name)
        self.logger.profiling_start('modelrun.run', self.name)

        if self.status == 'Built':
            if not self.model_horizon:
                raise SmifModelRunError("No timesteps specified for model run")
            if warm_start_timestep:
                idx = self.model_horizon.index(warm_start_timestep)
                self.model_horizon = self.model_horizon[idx:]
            self.status = 'Running'
            modelrunner = ModelRunner()
            modelrunner.solve_model(self, store)
            self.status = 'Successful'
        else:
            raise SmifModelRunError("Model is not yet built.")

        self.logger.profiling_stop('modelrun.run', self.name)


class ModelRunner(object):
    """The ModelRunner orchestrates the simulation of a SoSModel over decision iterations and
    timesteps as provided by a DecisionManager.
    """
    def __init__(self):
        self.logger = getLogger(__name__)

    def solve_model(self, model_run, store):
        """Solve a ModelRun

        This method steps through the model horizon, building
        a job graph and submitting this to the scheduler
        at each decision loop.

        Arguments
        ---------
        model_run : :class:`smif.controller.modelrun.ModelRun`
        store : :class:`smif.data_layer.Store`
        """
        # Solve the model run: decision loop generates a series of bundles of independent
        # decision iterations, each with a number of timesteps to run
        self.logger.debug("Solving the models over all timesteps: %s", model_run.model_horizon)

        # Initialise the decision manager (and hence decision modules)
        self.logger.debug("Initialising the decision manager")
        decision_manager = DecisionManager(store,
                                           model_run.model_horizon,
                                           model_run.name,
                                           model_run.sos_model)

        # Initialise the job scheduler
        self.logger.debug("Initialising the job scheduler")
        job_scheduler = JobScheduler()
        job_scheduler.store = store

        for bundle in decision_manager.decision_loop():
            # each iteration is independent at this point, so the following loop is a
            # candidate for running in parallel
            job_graph = self.build_job_graph(model_run, bundle)

            job_id, err = job_scheduler.add(job_graph)
            self.logger.debug("Running job %s", job_id)
            if err is not None:
                status = job_scheduler.get_status(job_id)
                self.logger.debug("Job %s %s", job_id, status['status'])
                raise err

    def build_job_graph(self, model_run, bundle):
        """ Build a job graph

        Build and return the job graph for an entire bundle, including before_model_run jobs
        when the models were not yet initialised.

        Constraints:
        - Bundle must have keys: 'decision_iterations' and 'timesteps'
        - Running a bundle runs each (decision iteration, timestep) pair specified by the
          combinations of decision iterations and timesteps
        - (decision iteration, timestep) pairs must be unique over an entire model run
        - In a single bundle, timesteps must be a consecutive subset of the model horizon
          timesteps

        The first timestep in each decision iteration of a bundle is either:
        - the first timestep in the model horizon and initialised from the model run starting
          point with scenario data and initial-timestep interventions only
        - or another timestep, picking up from where some previous (timestep, decision
          iteration) left off, which is explicitly included in the bundle.

        If a bundle's timesteps start from a timestep after the first one in the model horizon,
        the bundle must provide 'decision_links', and bundle must start from the very next
        timestep available in the model horizon.

        Jobs need to be able to identify a point to pick up from, namely the (timestep,
        decision iteration) which identifies the immediately preceding simulation state.

        E.g. request running first two timesteps::

            {
                'decision_iterations': [0, 1],
                'timesteps': [0, 1]
            }

        Run first two timesteps again, with an updated decision::

            {
                'decision_iterations': [1, 2],
                'timesteps': [0, 1]
            }

        Results meet decision requirements, so run next two timesteps, linking this bundle's
        decision iterations to previous decision iterations::

            {
                'decision_iterations': [3, 4],
                'timesteps': [2, 3],
                'decision_links': {3: 1, 4: 2}
            }


        Arguments
        ---------
        model_run: :class:`smif.controller.modelrun.ModelRun` bundle: :class:`dict`

        Returns
        -------
        :class:`networkx.Graph` A populated job graph with edges showing dependencies between
            different operations and timesteps
        """
        job_graph = nx.DiGraph()

        # Solve the model run: decision loop generates a series of bundles of independent
        # decision iterations, each with a number of timesteps to run
        for decision_iteration in bundle['decision_iterations']:
            self.logger.info('Running decision iteration %s', decision_iteration)

            for timestep_index, timestep in enumerate(bundle['timesteps']):
                self.logger.info('Running timestep %s', timestep)
                # one simulate job node per model
                job_graph.add_nodes_from(
                    self._make_simulate_job_nodes(
                        model_run.name,
                        model_run.sos_model.models,
                        decision_iteration,
                        timestep,
                        model_run.model_horizon
                    )
                )
                # edges to match within-timestep dependencies
                job_graph.add_edges_from(
                    self._make_current_simulate_job_edges(
                        model_run.name,
                        model_run.sos_model.dependencies,
                        timestep,
                        decision_iteration
                    )
                )

                # connect any between-timestep dependencies
                if timestep_index == 0:
                    # first timestep in bundle
                    try:
                        # connect to outputs from a previous bundle
                        relative = RelativeTimestep.PREVIOUS
                        previous_timestep = relative.resolve_relative_to(
                            timestep, model_run.model_horizon)
                        previous_decision_iteration = \
                            bundle['decision_links'][decision_iteration]
                        job_graph.add_edges_from(
                            self._make_between_bundle_previous_simulate_job_edges(
                                model_run.name,
                                model_run.sos_model.dependencies,
                                timestep,
                                previous_timestep,
                                decision_iteration,
                                previous_decision_iteration
                            )
                        )
                    except SmifTimestepResolutionError:
                        # no previous timestep, use scenarios to provide initial intertimestep
                        # dependenciess
                        job_graph.add_edges_from(
                            self._make_initial_previous_simulate_job_edges(
                                model_run.name,
                                model_run.sos_model.dependencies,
                                timestep,
                                decision_iteration
                            )
                        )

                else:
                    # subsequent timestep in a bundle - connect to previous timestep
                    previous_timestep = bundle['timesteps'][timestep_index - 1]
                    job_graph.add_edges_from(
                        self._make_within_bundle_previous_simulate_job_edges(
                            model_run.name,
                            model_run.sos_model.dependencies,
                            timestep,
                            previous_timestep,
                            decision_iteration
                        )
                    )

        if not model_run.initialised:
            # one before_model_run job per model
            self.logger.info("Initialising each of the sector models")
            job_graph.add_nodes_from(
                self._make_before_model_run_job_nodes(
                    model_run.name,
                    model_run.sos_model.models,
                    model_run.model_horizon
                )
            )
            # must run before any simulate jobs
            for decision_iteration in bundle['decision_iterations']:
                for timestep in bundle['timesteps']:
                    job_graph.add_edges_from(
                        self._make_before_model_run_job_edges(
                            model_run.name,
                            model_run.sos_model.models,
                            timestep,
                            decision_iteration
                        )
                    )
            model_run.initialised = True

        if not nx.is_directed_acyclic_graph(job_graph):
            raise NotImplementedError(
                "SosModel dependency graphs must not contain within-timestep cycles")

        return job_graph

    @staticmethod
    def _make_before_model_run_job_nodes(modelrun_name, models, horizon):
        return [
            (
                ModelRunner._make_job_id(
                    modelrun_name, model.name, ModelOperation.BEFORE_MODEL_RUN),
                {
                    'model': model,
                    'modelrun_name': modelrun_name,
                    'current_timestep': None,
                    'timesteps': horizon,
                    'decision_iteration': None,
                    'operation': ModelOperation.BEFORE_MODEL_RUN
                }
            )
            for model in models
        ]

    @staticmethod
    def _make_before_model_run_job_edges(modelrun_name, models, timestep, decision_iteration):
        edges = []
        for model in models:
            from_id = ModelRunner._make_job_id(
                modelrun_name, model.name, ModelOperation.BEFORE_MODEL_RUN)
            to_id = ModelRunner._make_job_id(
                modelrun_name, model.name, ModelOperation.SIMULATE, timestep,
                decision_iteration)
            edges.append((from_id, to_id))
        return edges

    @staticmethod
    def _make_simulate_job_nodes(modelrun_name, models, decision_iteration, timestep, horizon):
        return [
            (
                ModelRunner._make_job_id(
                    modelrun_name, model.name, ModelOperation.SIMULATE, timestep,
                    decision_iteration),
                {
                    'model': model,
                    'modelrun_name': modelrun_name,
                    'current_timestep': timestep,
                    'timesteps': horizon,
                    'decision_iteration': decision_iteration,
                    'operation': ModelOperation.SIMULATE
                }
            )
            for model in models
        ]

    @staticmethod
    def _make_current_simulate_job_edges(modelrun_name, dependencies, timestep,
                                         decision_iteration):
        edges = []
        for dependency in dependencies:
            if dependency.timestep != RelativeTimestep.PREVIOUS:
                from_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.source_model.name, ModelOperation.SIMULATE,
                    timestep, decision_iteration)
                to_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.sink_model.name, ModelOperation.SIMULATE,
                    timestep, decision_iteration)
                edges.append((from_id, to_id))
        return edges

    @staticmethod
    def _make_within_bundle_previous_simulate_job_edges(modelrun_name, dependencies, timestep,
                                                        previous_timestep, decision_iteration):
        edges = []
        for dependency in dependencies:
            if dependency.timestep == RelativeTimestep.PREVIOUS:
                from_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.source_model.name, ModelOperation.SIMULATE,
                    previous_timestep, decision_iteration)
                to_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.sink_model.name, ModelOperation.SIMULATE,
                    timestep, decision_iteration)
                edges.append((from_id, to_id))
        return edges

    @staticmethod
    def _make_between_bundle_previous_simulate_job_edges(modelrun_name, dependencies, timestep,
                                                         previous_timestep, decision_iteration,
                                                         previous_decision_iteration):
        edges = []
        for dependency in dependencies:
            if dependency.timestep == RelativeTimestep.PREVIOUS:
                from_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.source_model.name, ModelOperation.SIMULATE,
                    previous_timestep, previous_decision_iteration)
                to_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.sink_model.name, ModelOperation.SIMULATE,
                    timestep, decision_iteration)
                edges.append((from_id, to_id))
        return edges

    @staticmethod
    def _make_initial_previous_simulate_job_edges(modelrun_name, dependencies, timestep,
                                                  decision_iteration):
        edges = []
        for dependency in dependencies:
            if isinstance(dependency.source_model, ScenarioModel):
                from_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.source_model.name, ModelOperation.SIMULATE,
                    timestep, decision_iteration)
                to_id = ModelRunner._make_job_id(
                    modelrun_name, dependency.sink_model.name, ModelOperation.SIMULATE,
                    timestep, decision_iteration)
                edges.append((from_id, to_id))
        return edges

    @staticmethod
    def _make_job_id(modelrun_name, model_name, operation, timestep=None,
                     decision_iteration=None):
        if operation == ModelOperation.BEFORE_MODEL_RUN:
            id_ = '%s_%s_%s' % (modelrun_name, operation.value, model_name)
        else:
            id_ = '%s_%s_%s_%s_%s' % (
                modelrun_name, operation.value, timestep, decision_iteration, model_name)
        return id_


class ModelRunBuilder(object):
    """Builds the ModelRun object from the configuration
    """
    def __init__(self):
        self.model_run = ModelRun()
        self.logger = getLogger(__name__)

    def construct(self, model_run_config):
        """Set up the whole ModelRun

        Arguments
        ---------
        model_run_config : dict
            A valid model run configuration dictionary
        """
        self.model_run.name = model_run_config['name']
        self.model_run.description = model_run_config['description']
        self.model_run.timestamp = model_run_config['stamp']
        self.model_run.initialised = False
        self._add_timesteps(model_run_config['timesteps'])
        self._add_sos_model(model_run_config['sos_model'])
        self._add_scenarios(model_run_config['scenarios'])
        self._add_narratives(model_run_config['narratives'])
        self._add_strategies(model_run_config['strategies'])

        self.model_run.status = 'Built'

    def validate(self):
        """Check and/or assert that the modelrun is correctly set up
        - should raise errors if invalid
        """
        assert self.model_run is not None, "Sector model not loaded"
        self.model_run.validate()
        return True

    def finish(self):
        """Returns a configured model run ready for operation

        """
        if self.model_run.status == 'Built':
            self.validate()
            return self.model_run
        else:
            raise RuntimeError("Run construct() method before finish().")

    def _add_sos_model(self, sos_model_object):
        """

        Arguments
        ---------
        sos_model_object : smif.model.sos_model.SosModel
        """
        self.model_run.sos_model = sos_model_object

    def _add_timesteps(self, timesteps):
        """Set the timesteps of the system-of-systems model

        Arguments
        ---------
        timesteps : list
            A list of timesteps
        """
        self.logger.info("Adding timesteps to model run")
        self.model_run.model_horizon = timesteps

    def _add_scenarios(self, scenarios):
        """

        Arguments
        ---------
        scenarios : dict
            A dictionary of {scenario set: scenario name}, one for each scenario set
        """
        self.model_run.scenarios = scenarios

    def _add_narratives(self, narratives):
        """

        Arguments
        ---------
        narratives : list
            A list of smif.parameters.Narrative objects
        """
        self.model_run.narratives = narratives

    def _add_strategies(self, strategies):
        """

        Arguments
        ---------
        narratives : list
            A list of strategies
        """
        self.model_run.strategies = strategies



"""Each dimension of a multi-dimensional dataset can be indexed by a set of coordinates.

Dimensions might be represented as Euclidean coordinates over a grid::

    >>> x = Coordinates('x', range(0, 100))
    >>> y = Coordinates('y', range(0, 100))

A subset of Local Authority Districts in England would be a region-based spatial dimension::

    >>> local_authority_districts = Coordinates('LADs', ['E07000128', 'E07000180'])

The hours of an average annual day would be a temporal dimension::

    >>> hours = Coordinates('daily_hours', [
    ...     {'name': 0, 'interval': [['PT0H', 'PT1H']]}
    ...     {'name': 1, 'interval': [['PT1H', 'PT2H']]}
    ...     {'name': 2, 'interval': [['PT2H', 'PT3H']]}
    ... ])

The economic sectors from the International Standard Industrial Classification of All Economic
Activities (ISIC), revision 4 would be a categorical dimension::

    >>> economic_sector = Coordinates('ISICrev4', [
    ...     {'name': 'A', 'desc': 'Agriculture, forestry and fishing'},
    ...     {'name': 'B', 'desc': 'Mining and quarrying'},
    ...     {'name': 'C', 'desc': 'Manufacturing'},
    ...     {'name': 'D', 'desc': 'Electricity, gas, steam and air conditioning supply'}
    ... ])

"""


class Coordinates(object):
    """Coordinates provide the labels to index a dimension, along with metadata that may be
    useful to describe, visualise or convert between dimensions.

    Coordinate element names are used to label each position along a finite dimension.

    A dict mapping dimension name to list of coordinate elements can be passed to a
    :class:`~smif.metadata.spec.Spec` (or :class:`xarray.DataArray`) as `coords`.

    Attributes
    ----------
    name : str
        Dimension name
    dim : str
        Alias for dimension name
    ids : list
        List of labels
    elements : list[dict]
        List of labels with metadata

    Parameters
    ----------
    name : str
        Name to identify the dimension that these coordinates index
    elements : list
        List of simple data types (used to identify elements), or a list of dicts with 'name'
        key and other metadata

    Raises
    ------
    ValueError
        If the list of elements is empty, or infinite
    KeyError
        If the elements are not a list of simple data types
        or a list of dicts with a 'name' key
    """
    def __init__(self, name, elements):
        self.name = name
        self._ids = None
        self._elements = None
        self._set_elements(elements)

    def __eq__(self, other):
        return self.name == other.name \
            and self.elements == other.elements

    def __hash__(self):
        return hash(tuple(frozenset(e.items()) for e in self._elements))

    def __repr__(self):
        return "<Coordinates name='{}' elements={}>".format(self.name, self.ids)

    @property
    def elements(self):
        """Elements are a list of dicts with at least a 'name' key

        Coordinate elements should not be changed.
        """
        return self._elements

    @property
    def ids(self):
        """Element ids is a list of coordinate identifiers
        """
        return self._ids

    @property
    def names(self):
        """Names is an alias for Coordinates.ids
        """
        return self._ids

    def _set_elements(self, elements):
        """Set elements with a list of ids (string or numeric) or dicts (including key 'id')
        """
        if not elements:
            raise ValueError("Coordinates.elements must not be empty")

        try:
            len(elements)
        except TypeError:
            raise ValueError("Coordinate.elements must be finite in length")

        try:
            self._ids = [e['name'] for e in elements]
            self._elements = elements
        except KeyError:
            # elements must have name
            msg = "Elements in dimension '{}' must have a name field, " \
                  "or be a simple list of identifiers"
            raise KeyError(msg.format(self.name))
        except TypeError:
            # elements might not be dict-like - in which case, treat them as names
            self._ids = elements
            self._elements = [{"name": e} for e in elements]

    @property
    def dim(self):
        """Dim (dimension) is an alias for Coordinates.name
        """
        return self.name

    @dim.setter
    def dim(self, dim):
        """Set name as dim
        """
        self.name = dim
